


proc InitGIDProject { dir } {
    global MenuNames MenuNamesP MenuEntries MenuCommands MenuAcceler 
    global MenuEntriesP MenuEntriesP MenuAccelerP

    if { [catch { set HelpMenuName [_ Help] }] } {
	set HelpMenuName Help
    }
    set ipos [lsearch $MenuNames $HelpMenuName]

    if { $ipos != -1 } {
	set MenuEntries($ipos) [linsert $MenuEntries($ipos) 0 "Abaqus interface help" --- ]
	set MenuCommands($ipos) [linsert $MenuCommands($ipos) 0 \
		                     [list -np- HelpWindow "CUSTOM_HELP_REFERENCE" \
		                          [file join $dir info abaqus_toc.html] \
		                          [file join $dir info abaqus.html]] \
		                     ""]
	set MenuAcceler($ipos) [linsert $MenuAcceler($ipos) 0 "" ""]
    }

    set ipos [lsearch $MenuNamesP $HelpMenuName]
    if { $ipos != -1 } {
	set MenuEntriesP($ipos) [linsert $MenuEntries($ipos) 0 "Abaqus interface help" --- ]
	set MenuCommandsP($ipos) [linsert $MenuCommands($ipos) 0 \
		                     [list -np- HelpWindow "CUSTOM_HELP_REFERENCE" \
		                          [file join $dir info abaqus_toc.html] \
		                          [file join $dir info abaqus.html]] \
		                     ""]
	set MenuAccelerP($ipos) [linsert $MenuAcceler($ipos) 0 "" ""]

    }
    CreateTopMenus
}

namespace eval abaqus {}

proc abaqus::writefile {} {
    set err [catch { abaqus::_writefile } string]
    if { $err } { WarnWin $string }
    return $string
}

proc abaqus::_writefile {} {
    set _ ""
    append _ "*HEADING\nABAQUS job created on [clock format [clock seconds] -format %d-%h-%y] at "
    append _ "[clock format [clock seconds] -format %H:%M:%S]\n"
    append _ "*NODE\n"
    foreach "n x y z" [.central.s info Mesh Nodes] {
	append _ [format " %d, %.6g, %.6g, %.6g\n" $n $x $y $z]
    }
    set IsQuadratic [lindex [.central.s info Project] 5]
    foreach "elemtype name" [list Linear *** Triangle *** Quadrilateral R3D4 Tetrahedra *** \
		                 Hexahedra C3D8R] {
	set elems [.central.s info Mesh Elements $elemtype]
	if { [llength $elems] } {
	    switch $elemtype {
		Linear { switch $IsQuadratic 0 { set n 2 } default { set n 3 } }
		Triangle { switch $IsQuadratic 0 { set n 3 } default { set n 6 } }
		Quadrilateral { switch $IsQuadratic 0 { set n 4 } 1 { set n 8 } 2 { set n 9 } }
		Tetrahedra { 
		    switch $IsQuadratic {
		        0 { set n 4 }
		        default { set n 10 ; set name C3D10 }
		    }
		}
		Hexahedra { switch $IsQuadratic 0 { set n 8 } 1 { set n 20 } 2 { set n 27 } }
	    }
	    append _ "*ELEMENT, TYPE=$name\n"
	    set len [llength $elems]
	    set incr [expr {$n+2}]
	    for { set i 0 } { $i < $len } { incr i $incr } {
		append _ "[join [lrange $elems $i [expr {$i+$n}]] {,   }]\n"
	    }
	}
    }
    foreach cndname [list Point_Nset Line_Nset Surface_Nset Volume_Nset] {
	foreach cnd [.central.s info conditions $cndname mesh] {
	    lappend nset([lindex $cnd 3]) [lindex $cnd 1]
	}
    }
    foreach i [array names nset] {
	set nodes [lsort -integer -unique $nset($i)]
	append _ "*NSET, NSET=$i\n"
	set inum 0
	foreach j $nodes {
	    append _ " $j,"
	    incr inum
	    if { $inum == 8 } {
		set inum 0
		append _ "\n"
	    }
	}
	set _ [string trimright $_ ",\n"]
	append _ "\n"
    }
    foreach cndname [list Line_Elset Surface_Elset Volume_Elset] {
	foreach cnd [.central.s info conditions $cndname mesh] {
	    lappend elset([lindex $cnd 3]) [lindex $cnd 1]
	}
    }
    foreach i [array names elset] {
	set nodes [lsort -integer -unique $elset($i)]
	append _ "*ELSET, ELSET=$i\n"
	set inum 0
	foreach j $nodes {
	    append _ " $j,"
	    incr inum
	    if { $inum == 8 } {
		set inum 0
		append _ "\n"
	    }
	}
	set _ [string trimright $_ ",\n"]
	append _ "\n"
    }
    foreach cnd [.central.s info conditions Surface_Surfaceset mesh] {
	lappend surfaceset([lindex $cnd 3]) [lrange $cnd 0 1]
    }
    foreach i [array names surfaceset] {
	set elems [lsort -integer -unique -index 0 $surfaceset($i)]
	append _ "*surface, type=ELEMENT,name=$i\n"   
	set inum 0
	foreach j $elems {
	    append _ "[lindex $j 0],S[lindex $j 1]\n"
	}
    }

# Tetrahedral element faces

# Face 1    1 2 3 face 
# Face 2    1 4 2 face 
# Face 3    2 4 3 face 
# Face 4    3 4 1 face 

# Hexahedron (brick) element faces

# Face 1    1 2 3 4 face 
# Face 2    5 8 7 6 face 
# Face 3    1 5 6 2 face 
# Face 4    2 6 7 3 face 
# Face 5    3 7 8 4 face 
# Face 6    4 8 5 1 face 


    return $_
}

proc abaqus::ComunicateWithGiD { condname op args } {
    variable SetName

    switch $op {
	"INIT" {
	    set PARENT [lindex $args 0]
	    upvar      [lindex $args 1] ROW
	    set GDN [lindex $args 2]
	    set STRUCT [lindex $args 3]

	    set f [frame $PARENT.f]

	    set values ""
	    foreach cnd [.central.s info conditions $condname geometry] {
		set name [lindex $cnd 3]
		if { [lsearch $values $name] == -1 } { lappend values $name }
	    }
	    lappend values --- Set1 Set2 Set3 Set4

	    label $f.l1 -text [= Name]
	    ComboBox $f.cb1 -textvariable abaqus::SetName -values $values

	    grid $f.l1 $f.cb1 -sticky nw
	    grid configure $f.cb1 -sticky new -padx "1 5"
	    grid rowconfigure $f 1 -weight 1
	    grid columnconfigure $f 1 -weight 1

	    GidHelpRecursive $f [= "Enter the name for the set"]
	    grid $f -row $ROW -column 0 -sticky nsew -columnspan 2 -pady 3 -padx 2
	    grid rowconf $PARENT $ROW -weight 1
	    grid columnconf $PARENT 1 -weight 1
	    return ""
	}
	"SYNC" {
	    set GDN [lindex $args 0]
	    set STRUCT [lindex $args 1]

	    set SetName [string trim $SetName]
	    if { $SetName eq "" } {
		return [list ERROR [= "Name is not valid"]]
	    } else {
		DWLocalSetValue $GDN $STRUCT Name $SetName
	    }
	    return ""
	}
    }
}
