
package require tablelist
package require snit
package require icons
package require ico
package require Img
package require treectrl

################################################################################
#    text_and_variable
################################################################################

snit::widgetadaptor text_and_variable {
    option -variable ""

    delegate method * to hull
    delegate method _insert to hull as insert
    delegate method _delete to hull as delete
    delegate option * to hull
    
    variable updating 0

    constructor args {
	installhull [text $self]
	
	bind $win <Tab> "[bind all <Tab>] ; break"
	bind $win <<PrevWindow>> "[bind all <<PrevWindow>>] ; break"

	$self configurelist $args
    }
    destructor {
	$self clear
    }
    onconfigure -variable {value} {
	$self clear
	set options(-variable) $value
	if { $value ne "" } {
	    #ind $win <<Modified>> [mymethod modifytext]
	    trace add variable $value write "[mymethod modifyvariable];#"
	}
    }
    method clear {} {
	if { $options(-variable) ne "" } {
	    trace remove variable $options(-variable) write \
		"[mymethod modifyvariable];#"
	    #bind $win <<Modified>> ""
	}
    }
    method modifytext {} {
	if { $options(-variable) eq "" } { return }
	if { $updating } { return }
	set updating 1
	upvar #0 $options(-variable) v
	if { ![info exists v] } { set v "" }
	set data [$self get 1.0 end-1c]
	if { $data ne $v} { set v $data }
	$self edit modified 0
	set updating 0
    }
    method modifyvariable {} {
	if { $updating } { return }
	set updating 1
	upvar #0 $options(-variable) v
	if { ![info exists v] } { set v "" }
	set data [$self get 1.0 end-1c]
	if { $data ne $v } {
	    $self delete 1.0 end
	    $self insert end $v
	}
	$self edit modified 0
	set updating 0
    }
    method insert { args } {
	set ret [eval $self _insert $args]
	$self modifytext
	return $ret
    }
    method delete { args } {
	set ret [eval $self _delete $args]
	$self modifytext
	return $ret
    }

}

################################################################################
#    ComboBox_and_valuesvar
################################################################################

snit::widgetadaptor ComboBox_and_valuesvar {
    option -valuesvariable ""
    option -values ""

    delegate method * to hull
    delegate option * to hull
    delegate option -_values to hull as -values

    constructor args {
	installhull [ComboBox $self]
	$self configurelist $args
    }
    destructor {
	$self clear
    }
    onconfigure -valuesvariable {value} {
	if { $options(-valuesvariable) ne "" } {
	    trace remove variable $options(-valuesvariable) write \
		"[mymethod modifyvariable];#"
	}
	set options(-valuesvariable) $value
	if { $value ne "" } {
	    trace add variable $value write "[mymethod modifyvariable];#"
	}
    }
    onconfigure -values {value} {
	upvar #0 $options(-valuesvariable) v
	if { ![info exists v] } { set v "" }
	set data [$self cget -values]
	if { $data ne $v} {
	    $self configure -_values $v
	}
    }
    method clear {} {
	if { $options(-valuesvariable) ne "" } {
	    trace remove variable $options(-valuesvariable) write \
		"[mymethod modifyvariable];#"
	}
    }
    method modifyvariable {} {
	upvar #0 $options(-valuesvariable) v
	if { ![info exists v] } { set v "" }
	set data [$self cget -values]
	if { $data ne $v} {
	    $self configure -_values $v
	}
    }
}

################################################################################
#     icon_chooser
################################################################################


namespace eval icon_chooser {
    variable default_library ""
    variable imported_images ""
}

proc icon_chooser::giveimage { icodesc } {
    variable imported_images

    foreach "ico file" $icodesc break
    regsub -all { } $ico {_} ico ;#avoid names with spaces (problems using tile)
    if { $file eq "" } {
	set name [icons::icons create $ico]
	if { $name eq "" } {
	    error "icon '$icodesc' cannot be found"
	}
	return ::icon::$name
    } elseif { $file eq "imported_images" } {
	foreach i $imported_images {
	    if { $ico eq [lindex $i 0] } { return [lindex $i 1] }
	}
    } elseif { [regexp {(?i)^\.(dll|ico|exe)$} [file extension $file]] } {
	return [::ico::getIcon $file $ico]
    } else {
	set name [icons::icons create -file $file $ico]
	if { $name eq "" } {
	    error "icon '$icodesc' cannot be found"
	}
	return ::icon::$name
    }
    error "error in icon_chooser::giveimage icodesc=$icodesc"
}

proc icon_chooser::choose_icon { parent } {
    variable default_library

    set w $parent.%AUTO%
    set w [dialogwin_snit $w -title [_ "Choose icon"] -morebuttons [list [_ "Clear"]]]
    set f [$w giveframe]
    
    set values [list tkIcons [_ "Imported images"]]
    if { $default_library eq "" } {
	set default_library [lindex $values 0]
    }
    if { [info exists ::env(WINDIR)] && \
	[file exists [file join $::env(WINDIR) system32 shell32.dll]] } {
	lappend values [file join $::env(WINDIR) system32 shell32.dll]
    }
    ComboBox $f.cb1 -textvariable [$w give_uservar file] -width 10 \
	-values $values -modifycmd [namespace code [list _fill_canvas $f.s.c $w]] \
	-editable 0
    Button $f.b1 -image [giveimage fileopen-16] -relief link -helptext \
	[_ "Select icon library"] -command [namespace code \
	    [list _select_library $f.s.c $w]] \
	-bd 1
    frame $f.f1
    radiobutton $f.f1.r1 -text 16 -variable [$w give_uservar width] -value 16 \
	-command [namespace code [list _fill_canvas $f.s.c $w]]
    radiobutton $f.f1.r2 -text 22-24 -variable [$w give_uservar width] -value 22 \
	-command [namespace code [list _fill_canvas $f.s.c $w]]
    button $f.f1.b1 -text [_ "Import..."] -padx 5 -command [namespace code \
	    [list _import_image $f.s.c $w]]
    button $f.f1.b2 -text [_ "Delete"] -padx 5 -command [namespace code \
	    [list _delete_imported_image $f.s.c $w]]
    grid $f.f1.r1 $f.f1.r2 $f.f1.b1 $f.f1.b2 -padx 5
    grid remove $f.f1.b1 $f.f1.b2
    
    ScrolledWindow $f.s -relief sunken -borderwidth 0
    canvas $f.s.c -bg white -bd 0 -highlightthickness 1 \
	-width 224 -height 224 -takefocus 1
    $f.s setwidget $f.s.c

    grid $f.cb1 $f.b1 -sticky nw -padx 3 -pady 3
    grid $f.f1    -   -sticky nw
    grid $f.s     -   -sticky nsew -padx 1 -pady 1
    grid configure $f.cb1 -sticky nwe
    grid columnconfigure $f 0 -weight 1
    grid rowconfigure $f 2 -weight 1

    $w set_uservar_value file $default_library
    $w set_uservar_value width 22
    $w set_uservar_value icondesc ""
    
    
    after 100 [namespace code [list _fill_canvas $f.s.c $w]]

    $f.cb1 bind <Return> [namespace code [list _fill_canvas $f.s.c $w]]
    bind $f.s.c <1> [list focus $f.s.c]
    bind $f.s.c <MouseWheel> {
	%W yview scroll [expr {- (%D / 120) * 4}] units
    }

    $w focusok
    set action [$w createwindow]
    while 1 {
	switch -- $action {
	    -1 - 0 {
		destroy $w
		return -
	    }
	    1 {
		set icondesc [$w give_uservar_value icondesc]
		if { $icondesc eq "" } {
		    WarnWin [_ "Select one icon"] $w
		} else {
		    set file [$w give_uservar_value file]
		    destroy $w
		    if { $file eq "tkIcons" } {
		        return [list $icondesc]
		    } elseif { $file eq [_ "Imported images"] } {
		        return [list $icondesc imported_images]
		    } else {
		        return [list $icondesc $file]
		    }
		}
	    }
	    2 {
		destroy $w
		return ""
	    }
	}
	set action [$w waitforwindow]
    }
}

proc icon_chooser::SaveImages { fout icondescs } {
    variable imported_images
    variable ImportedDates

    set names ""
    foreach icodesc $icondescs {
	foreach "ico file" $icodesc break
	regsub -all { } $ico {_} ico ;#avoid names with spaces (problems using tile)
	if { $file eq "imported_images" } {
	    lappend names $ico
	}
    }
    puts $fout "\n"
    foreach i $imported_images {
	foreach "name img date" $i break
	regsub -all { } $name {_} name ;#avoid names with spaces (problems using tile)
	if { $names ne "" && [lsearch -exact $names $name] == -1 } { continue }
	puts $fout "[list set ::icon_chooser::ImportedDates($name) $date]\n"
	puts $fout "[list image create photo ::icon_chooser::images::$name] -data \{"
	if { [regexp {(?i)\.jpe?g\d*$} $name] } {
	    set format jpeg
	} elseif { [regexp {(?i)\.png\d*$} $name] } {
	    set format png
	}  else { 
	    set format gif
	}
	puts $fout [$img data -format $format]
	puts $fout "\}\n"
    }
}

proc icon_chooser::AfterReadMacros {} {
    variable imported_images
    variable ImportedDates

    set imported_images ""
    
    set now [clock format [clock seconds] -format "%Y-%m-%d %H:%M:%S"]
    foreach img [info commands ::icon_chooser::images::*] {
	set name [namespace tail $img]        
	regsub -all { } $name {_} name ;#avoid names with spaces (problems using tile)
	if { [info exists ImportedDates($name)] } {
	    set data $ImportedDates($name)
	} else { set date $now }        
	lappend imported_images [list $name $img $ImportedDates($name)]
    }
    set imported_images [lsort -index 2 $imported_images]
}


	

proc icon_chooser::_delete_imported_image { c w } {
    variable imported_images

    set ic [$w give_uservar_value icondesc]
    if { $ic eq "" } { return }
    foreach tag [$c gettags A$ic] {
	if { [regexp {^A(.*)$} $tag {} name] } {
	    $w set_uservar_value icondesc $name
	    regsub -all { } $name {_} name ;#avoid names with spaces (problems using tile)
	}
	if { [regexp {^N(\d+)$} $tag {} number] } { break }
    }

    if { [::toolbarmacros::is_image_in_use $name] } {
	WarnWin [_ "this icon can't be deleted because it is used in the macros toolbar"]
	return 1
    }

    set idx 0
    foreach i $imported_images {
	if { $name eq [lindex $i 0] } {
	    image delete [lindex $i 1]
	    set imported_images [lreplace $imported_images $idx $idx]
	    unset icon_chooser::ImportedDates($name)
	    break
	}
	incr idx
    }
    toolbarmacros::SaveMacros
    _fill_canvas $c $w $number
}

proc icon_chooser::_import_image { c w } {
    variable imported_images

    set types [list \
	    [list [_ "Image files"] [list .gif .jpg .png]] \
	    [list [_ "All files"] .*]]
    
    set basefile [$w give_uservar_value file]
    if { $basefile ne "tkIcons" && $basefile ne [_ "Imported images"] } {
	set initialfile [file tail $basefile]
    } else {
	set initialfile ""
    }

    set file [Browser-ramR file read $w [_ "Select image"] \
	    $initialfile $types "" 0]
    if { $file == "" } { return }

    set idx ""
    while 1 {
	set name [file tail $file]$idx
	regsub -all { } $name {_} name ;#avoid names with spaces (problems using tile)
	set found 0
	foreach i $imported_images {
	    if { $name eq [lindex $i 0] } {
		set found 1
		break
	    }
	}
	if { !$found } { break }
	if { $idx eq "" } { set idx 1 } else { incr idx }
    }
    set err [catch {image create photo icon_chooser::images::$name -file $file} img]
    if { $err } {
	WarnWin [_ "Error importing image '%' (%s)" $file $img]
	return
    }
    set now [clock format [clock seconds] -format "%Y-%m-%d %H:%M:%S"]
    set ::icon_chooser::ImportedDates($name) $now
    lappend imported_images [list $name $img $now]
    toolbarmacros::SaveMacros
    _fill_canvas $c $w
}

proc icon_chooser::_select_library { c w } {

    set types [list \
	    [list [_ "Icons files"] [list .dll .exe .ico]] \
	    [list [_ "All files"] .*]]

    set initialfile [file tail [$w give_uservar_value file]]
    
    set file [Browser-ramR file read $w [_ "Select library"] \
	    $initialfile $types "" 0]

    if { $file == "" } { return }
    $w set_uservar_value file $file
    _fill_canvas $c $w
}

proc icon_chooser::DisplayHelp { c help what args } {
    variable displayhelpafterid

    if { ![winfo exists $c] } { return }
    if { [info exists displayhelpafterid] } {
	after cancel $displayhelpafterid
	unset displayhelpafterid
    }
    switch $what {
	enter {
	    set displayhelpafterid [after 1000 [namespace code \
		        [list DisplayHelp $c $help go]]]
	}
	leave {
	    # nothing
	}
	go {
	    destroy $c._help
	    set w [toplevel $c._help]
	    wm overrideredirect $w 1
	    set x [winfo pointerx .]
	    set y [winfo pointery .]
	    wm geometry $w +[expr {$x+8}]+[expr {$y-14}]
	    pack [label $w.l -text $help -relief groove -bg white -bd 1 \
		    -justify left]
	    bind $w <1> [list destroy $w]
	    bind $w <KeyPress> [list destroy $w]
	    bind $w <Motion> [namespace code \
		    [list DisplayHelp $c $help motion $x $y]]
	    focus $w
	    grab $w
	}
	motion {
	    foreach "x1 y1" $args break
	    set x2 [winfo pointerx .]
	    set y2 [winfo pointery .]
	    if { abs($x1-$x2)+abs($y1-$y2) > 4 } { destroy $c._help }
	}
    }
}

proc icon_chooser::_fill_canvas_select { c w what } {

    switch $what {
	cursor {
	    set id [$c find withtag current]
	}
	Home {
	    set id [$c find withtag N1]
	}
	End {
	    set id [$c find withtag N[$w give_uservar_value icon_number]]
	}
	Left - Right - Up - Down {
	    if { [$w give_uservar_value icondesc] eq "" } {
		return [_fill_canvas_select $c $w Home]
	    }
	    foreach tag [$c gettags A[$w give_uservar_value icondesc]] {
		if { [regexp {^N(\d+)$} $tag {} number] } { break }
	    }
	    switch $what {
		Left { 
		    if { $number > 1 } { incr number -1 }
		}
		Right { 
		    if { $number < [$w give_uservar_value icon_number] } { incr number }
		}
		Up {
		    if { $number > [$w give_uservar_value icon_number_row] } {
		        incr number -[$w give_uservar_value icon_number_row]
		    }
		}
		Down {
		    if { $number < [$w give_uservar_value icon_number]-\
		        [$w give_uservar_value icon_number_row]+1 } {
		        incr number [$w give_uservar_value icon_number_row]
		    }
		}
	    }
	    set id N$number
	}
    }
    _fill_canvas_select_do $c $w $id
}

proc icon_chooser::_center_canvas_in_id { canvas i } {
    foreach "x y" [$canvas coords $i] break
    foreach "- - width height" [$canvas cget -scrollregion] break
    
    foreach "f1 f2" [$canvas xview] break
    set xs [expr $x/double($width)-($f2-$f1)/2.0]
    if { $xs < 0 } { set xs 0 }
    $canvas xview moveto $xs
    
    foreach "f1 f2" [$canvas yview] break
    set ys [expr $y/double($height)-($f2-$f1)/2.0]
    if { $ys < 0 } { set ys 0 }
    $canvas yview moveto $ys
}

proc icon_chooser::_fill_canvas_select_do { c w id } {

    $c delete selmask
    foreach "x y" [$c coords $id] break
    foreach tag [$c gettags $id] {
	if { [regexp {^A(.*)$} $tag {} name] } {
	    $w set_uservar_value icondesc $name
	}
    }
    set width [image width [$c itemcget $id -image]]
    $c create rectangle [expr {$x-1}] [expr {$y-1}] [expr {$x+$width+2}] \
	[expr {$y+$width+2}] \
	-outline "" -fill grey60 -tags selmask
    $c lower selmask
    _center_canvas_in_id $c $id
}

proc icon_chooser::_fill_canvas { c w { sel_idx "" } } {

    $w configure -cursor watch
    update
    set err [catch { _fill_canvasDo $c $w $sel_idx }]
    $w configure -cursor ""
    if { $err } {
	WarnWin [_ "Library not valid"] $w
    }
}

proc icon_chooser::_fill_canvasDo { c w  { sel_idx "" } } {
    variable imported_images
    variable default_library

    set file [$w give_uservar_value file]
    set width [$w give_uservar_value width]
    $c delete all

    set f [$w giveframe]
    if { $file eq [_ "Imported images"] } {
	grid $f.f1.b1 $f.f1.b2 -pady 5
	$f.f1.r1 configure -state disabled
	$f.f1.r2 configure -state disabled
    } else {
	grid remove $f.f1.b1 $f.f1.b2
	$f.f1.r1 configure -state normal
	$f.f1.r2 configure -state normal
    }

    set l ""
    if { $file eq [_ "Imported images"] } {
	foreach i $imported_images {
	    foreach "name img date" $i break
	    regsub -all { } $name {_} name ;
	    if { [image width $img] > $width } {
		set width [image width $img]
	    }
	    lappend l $name $img "$name\n[_ {Creation date:}] $date"
	}
    } elseif { [regexp {(?i)^\.(dll|ico|exe)$} [file extension $file]] } {
	if { $width == 22 } { set width 24 }
	set idx 0
	foreach i [::ico::getIconList $file] {
	    foreach "widthi - -" $i break
	    if { $widthi == $width } {
		lappend l $idx [::ico::getIcon $file $idx] $idx
	    }
	    incr idx
	}
    } else {
	foreach i [icons::icons create -file $file -group $width] {
	     lappend l $i ::icon::$i $i
	}
    }
    foreach "x y idx idxrow" [list 4 4 0 0] break
    foreach "name image help" $l {
	set id [$c create image $x $y -image $image -anchor nw \
		-tags [list A$name icon N[incr idx]]]
	if { $sel_idx == $idx } {
	    _fill_canvas_select_do $c $w $id
	}
	incr x [expr {$width+4}]
	if { $x > 200 } {
	    if { !$idxrow } { set idxrow $idx }
	    set x 4
	    incr y [expr {$width+4}]
	}
	$c bind icon <1> [namespace code [list _fill_canvas_select $c $w cursor]]
	$c bind icon <Double-1> [list $w invokeok]
	bind $c <Return> [list $w invokeok]
	$c bind $id <Enter> [namespace code [list DisplayHelp $c $help enter]]
	$c bind $id <Leave> [namespace code [list DisplayHelp $c $help leave]]
    }
    foreach i [list Home End Left Right Up Down] {
	bind $c <$i> [namespace code [list _fill_canvas_select $c $w $i]]
    }
    if { $file eq [_ "Imported images"] } {
	bind $c <Delete> [namespace code [list _delete_imported_image \
		    $c $w]]
    }
    $w set_uservar_value icon_number_row $idxrow
    $w set_uservar_value icon_number $idx
    $w set_uservar_value icondesc ""
    $c configure -scrollregion [list 0 0 224 $y]
    set default_library $file
}

################################################################################
#    TreeCtrlList
################################################################################

::snit::widgetadaptor TreeCtrlList {
    option -columns

    delegate method * to hull
    delegate method _selection to hull as selection
    delegate option * to hull
    
    variable SortColumn
    variable itemStyle
    variable SystemButtonFace
    variable SystemHighlight
    variable SystemHighlightText
    
    constructor args {
	installhull using treectrl -highlightthickness 0 -borderwidth 0
	$self debug configure -enable no -display no
	
	if { ![info exists SystemButtonFace] } {
	    set w [listbox .listbox]
	    set SystemButtonFace [$w cget -highlightbackground]
	    set SystemHighlight [$w cget -selectbackground]
	    set SystemHighlightText [$w cget -selectforeground]
	    destroy $w
	}

	set height [font metrics [$self cget -font] -linespace]
	if {$height < 24} {
	    set height 24
	}
	$self configure -showroot no -showbuttons no -showlines no -itemheight $height \
	    -selectmode extended -xscrollincrement 20 \
	    -scrollmargin 16 -xscrolldelay "500 50" -yscrolldelay "500 50"

	$self element create e1 image
	$self element create e2 text -fill [list $SystemHighlightText {selected focus}] \
	    -lines 1
	$self element create e3 text -lines 1
	$self element create e4 rect -fill [list $SystemHighlight {selected focus} \
		gray {selected !focus}] -showfocus yes -open we
	$self element create e5 text -lines 1 -format " " -datatype string

	set S [$self style create imagetext -orient horizontal]
	$self style elements $S {e4 e1 e2}
	$self style layout $S e1 -expand ns
	$self style layout $S e2 -padx {4 0} -squeeze x -expand ns
	$self style layout $S e4 -union [list e2] -iexpand nswe -ipadx 2

	set S [$self style create text -orient horizontal]
	$self style elements $S {e4 e2}
	$self style layout $S e2 -padx {2 0} -squeeze x -expand ns
	$self style layout $S e4 -union [list e2] -iexpand nswe -ipadx 2

	set S [$self style create image -orient horizontal]
	$self style elements $S {e4 e1 e2 e5}
	$self style layout $S e1 -expand ns
	$self style layout $S e2 -padx 0 -squeeze x -expand ns
	$self style layout $S e4 -union [list e1] -iexpand nswe -ipadx 2
	$self style layout $S e5 -padx 0 -squeeze x -expand nswe
		
	$self notify install event Header
	$self notify install detail Header invoke
	
	$self notify install event Drag
	$self notify install detail Drag begin
	$self notify install detail Drag end
	$self notify install detail Drag receive
	
	$self notify install event Edit
	$self notify install detail Edit accept
	
	
	$self notify bind $self <Edit-accept> {
	    %T item text %I 0 %t
	}       
	set SortColumn 0
	$self notify bind $self <Header-invoke> [mymethod headerinvoke %C]
	
	$self configurelist $args
    }
    onconfigure -columns {value} {
	set options(-columns) $value

	set w0 [font measure [$self cget -font] 0]
	set itemStyle ""
	set idx 0
	foreach i [list edit sensitive dragimage] {
	    set ::TreeCtrl::Priv($i,$self) ""
	}
	foreach "width name justify type" $value {
	     $self column create -text $name -width [expr {$w0*$width}] \
		-tag $name -justify $justify
	    lappend itemStyle $idx $type
	    switch $type {
		imagetext {
		    #lappend ::TreeCtrl::Priv(edit,$self) [list e2 e3]
		    lappend ::TreeCtrl::Priv(sensitive,$self)  \
		            [list $idx $type e1 e2]
		    lappend ::TreeCtrl::Priv(dragimage,$self)  \
		            [list $idx $type e1 e2]
		}
		text {
		    #lappend ::TreeCtrl::Priv(edit,$self) [list e2 e3]
#                     lappend ::TreeCtrl::Priv(sensitive,$self)  \
#                             [list $idx $type e2]
		    lappend ::TreeCtrl::Priv(dragimage,$self) \
		            [list $idx $type e2]
		}
		image {
#                     lappend ::TreeCtrl::Priv(sensitive,$self)  \
#                             [list $idx $type e1]
		    lappend ::TreeCtrl::Priv(dragimage,$self) \
		            [list $idx $type e1]
		}
	    }
	    incr idx
	}
#        set ::TreeCtrl::Priv(edit,$self) [list e2 e3]
#         set ::TreeCtrl::Priv(edit,$self) {e2}
#         set ::TreeCtrl::Priv(sensitive,$self) [list \
#                 [list [lindex $value 1] [lindex $value 3] e1 e2] \
#             ]
#         set ::TreeCtrl::Priv(dragimage,$self) [list \
#                 [list [lindex $value 1] [lindex $value 3] e1 e2] \
#         ]
	bindtags $self [list $self TreeCtrlFileList TreeCtrl [winfo toplevel $self] all]
    }
    method insert { index list } {
	if { $index ne "end" } { error "only implemented index end in insert" }

	set item [$self item create]
	eval [list $self item style set $item] $itemStyle

	set desc ""
	set idx 0
	foreach "- - - type" $options(-columns) {
	    switch $type {
		imagetext {
		    foreach "img txt" [lindex $list $idx] break
		    # dirty trick to avoid a crash in tktreectrl
		    if { $txt eq "" } { set txt " " }
		    set selimg ${img}::selected
		    if { [lsearch [image names] $selimg] == -1 } {
		        image create photo $selimg
		        $selimg copy $img
		        imagetint $selimg $SystemHighlight 128
		    }
		    lappend desc [list [list e1 -image [list $selimg {selected} $img {}]] \
		            [list e2 -text $txt]]
		}
		text {
		    set txt [lindex $list $idx]
		    if { $txt eq "" } { set txt " " }
		    lappend desc [list [list e2 -text $txt]]
		}
		image {
		    foreach "img txt" [lindex $list $idx] break
		    if { $txt eq "" } { set txt " " }
		    lappend desc [list [list e1 -image $img] [list e5 -data $txt]]
		}
	    }
	    incr idx
	}        
	eval [list $self item complex $item] $desc
	$self item lastchild root $item
	set ::TreeCtrl::Priv(DirCnt,$self) $item
    }
    method headerinvoke { C } {
	if {$C == $SortColumn} {
	    if {[$self column cget $SortColumn -arrow] eq "down"} {
		set order -increasing
		set arrow up
	    } else {
		set order -decreasing
		set arrow down
	    }
	} else {
	    if {[$self column cget $SortColumn -arrow] eq "down"} {
		set order -decreasing
		set arrow down
	    } else {
		set order -increasing
		set arrow up
	    }
	    $self column configure $SortColumn -arrow none
	    set SortColumn $C
	}
	set type [lindex $options(-columns) [expr {$C*4+3}]]
	$self column configure $C -arrow $arrow
	if { $type eq "image" } {
	    $self item sort root $order -column $C -element e5 -dictionary
	} else {
	    $self item sort root $order -column $C -dictionary
	}
    }
    method sortcolumn {} {
	return $SortColumn
    }
    method sortorder {} {
	if {[$self column cget $SortColumn -arrow] eq "down"} {
	    return decreasing
	} else {
	    return increasing
	}
    }
    method sortbycolumn { col { sortorder -increasing } } {

	switch -- $sortorder {
	    "-increasing" { set arrow up }
	    "-decreasing" { set arrow down }
	}
	$self column configure $SortColumn -arrow none
	set SortColumn $col

	set type [lindex $options(-columns) [expr {$col*4+3}]]
	$self column configure $col -arrow $arrow
	if { $type eq "image" } {
	    $self item sort root $sortorder -column $col -element e5 -dictionary
	} else {
	    $self item sort root $sortorder -column $col -dictionary
	}
    }
    method get { first { last "" } } {
	set ret ""
	if { $last ne "" } {
	    for { set i $first } { $i <= $last } { incr i } {
		lappend itemList $i
	    }
	} else { set itemList $first }
	set idx 0
	foreach i $itemList {
	    set id [$self index "rnc $i 0"]
	    set retel ""
	    foreach "- - - type" $options(-columns) {
		switch $type {
		    imagetext - text {
		        lappend retel [$self item element cget $id $idx e2 -text]
		    }
		    image {
		        lappend retel [$self item element cget $id $idx e5 -data]
		    }
		}
		incr idx
	    }
	    lappend ret $retel
	}
	if { [llength $itemList] == 1 } {
	    return [lindex $ret 0]
	} else {
	    return $ret
	}
    }
    method selection_r { what args } {

	switch $what {
	    set - add {
		if { [llength $args] != 1 } {
		    error "error in TreeCtrlList selection"
		}              
		if { ![string is integer -strict [lindex $args 0]] } {
		    return [$self _selection add [lindex $args 0]]
		} else {
		    return [$self _selection add "rnc [lindex $args 0] 0"]
		}
	    }
	    includes {
		if { [llength $args] != 1 } {
		    error "error in TreeCtrlList selection"
		}
		if { ![string is integer -strict [lindex $args 0]] } {
		    return [$self _selection includes [lindex $args 0]]
		} else {
		    return [$self _selection includes "rnc [lindex $args 0] 0"]
		}
	    }
	    clear {
		if { [llength $args] < 0 || [llength $args] > 2 } {
		    error "error in TreeCtrlList selection"
		}
		if { [llength $args] == 0 } {
		    return [$self _selection clear]
		}
		if { [lindex $args 0] eq "last" } {
		    set first last
		} else { set first [$self index "rnc [lindex $args 0] 0"] }
		if { [llength $args] == 1 } {
		    return [$self _selection clear $first]
		}
		if { [lindex $args 1] eq "last" } {
		    set last last
		} else { set last [$self index "rnc [lindex $args 1] 0"] }
		return [$self _selection clear $first $last]
	    }
	    default {
		return [eval [list $self _selection $what] $args]
	    }
	}
	error "error in TreeCtrlList selection"
    }
    method curselection {} {
	set ret ""
	foreach i [$self _selection get] {
	    lappend ret [lindex [$self item rnc $i] 0]
	}
	return $ret
    }
    method delete { first { last "" } } {
	if { $first == 0 && $last eq "end" } {
	    $self item delete 0 last
	    return
	}
	if { $last ne "" } {
	    for { set i $first } { $i <= $last } { incr i } {
		lappend itemList $i
	    }
	} else { set itemList $first }
	foreach i $itemList {
	    $self item delete "rnc $i 0"
	}
    }
}


################################################################################
#    toolbarmacros
################################################################################

# Fields in macrosdata:
# Number
# Group
# Description
# Accelerators
# Icon
# InToolbar
# Active
# PrePost
# CreationDate
# ModificationDate
# Code

namespace eval toolbarmacros {
    variable _isinit 0
    variable images
    variable preferences
}

namespace eval toolbarmacros::macrospace {
    variable macrosdata
}

proc toolbarmacros::Init {} {
    variable _isinit
    variable preferences

    set _isinit 1

    set preferences(toolbar_one_col) 0

    ReadMacros
    Create
    _update_MacrosList .gid.tmacros
    ApplyMacrosBindings
}

proc toolbarmacros::Create { { type "DEFAULT INSIDETOP"} } {
    global GidPriv GiDPrivToolBar
    variable images
    variable _isinit

    if { !$_isinit} {
	::toolbarmacros::Init
    }
    create_images

    set title [_ "Macros toolbar"]
    set name MacrosToolbar
    set geomname PrePostMacrosToolbarWindowGeom
    set windowname PrePostMacrosToolbarWindow
    set w .gid.bitmapsMacrosToolbar

    DestroySlave $w
  
    AddNewToolbar $name $geomname toolbarmacros::Create $title

    if { [ info exist GidPriv($geomname)] } {
	if { [lindex $type 0] != "DEFAULT" } {
	    set GidPriv($geomname) "$type [lrange $GidPriv($geomname) 1 end]"
	}
	if { [lindex $GidPriv($geomname) 0] == "NONE" } {
	    return ""
	}
    } elseif { [lindex $type 0] == "DEFAULT" && [lindex $type 1] != "" } {
	set GidPriv($geomname) "[lindex $type 1] {} 1 toolbarmacros::Create"
    }

    if { [lindex [GiD_Info Project] 4] != "POSTUSE" } {
	set what PRE
    } else { set what POST }

    if { ![ info exist GidPriv($geomname)] || \
	    ![string match INSIDE* [lindex $GidPriv($geomname) 0]] } {
	InitWindow $w $title $geomname toolbarmacros::Create
	if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
	wm transient $w .gid
	catch { wm attributes $w -toolwindow 1 }
	$w configure -bg $GidPriv(Color,ToolbarBg)
	if { [ winfo width $w] > [ winfo height $w]} {
	    set orient horizontal
	} else {
	    set orient vertical
	}
    } else {
	set vv $GidPriv($geomname)
	#if { [winfo exists $w] } { DestroySlave $w }
	set GidPriv($geomname) $vv
	set GiDPrivToolBar($geomname) $w
	set GidPriv($windowname) $w

	frame $w -bd 0 -bg $GidPriv(Color,ToolbarBg)

	bind $w <Destroy> "+ unset GiDPrivToolBar($geomname)"
# 	if { [winfo exists .gid.central.wins] } {
# 	    pack forget .gid.central.wins
# 	} else {
# 	    pack forget .gid.central.s
# 	}
	regexp {([a-zA-Z]*)([0-9]*)} [lindex $GidPriv($geomname) 0] {} \
		where pos
      set widget ""
# 	if { $pos != "" } {
# 	    set widget [lindex [pack slaves .gid.central] $pos]
# 	} else { set widget "" }
	switch $where {
	    INSIDETOP {
# 		pack $w -in .gid.central -side top -anchor nw \
# 		        -fill both -expand 0 -padx 0 -pady 0
# 		if { [winfo exists .gid.central.wins] } {
# 		    pack .gid.central.wins -expand 1 -fill both -side top
# 		} else {
# 		    pack .gid.central.s -fill both -expand yes -side top
              #}
              GridInside $geomname $w .gid.top "we"
              set orient horizontal
	    }
	    INSIDELEFT {
# 		pack $w -in .gid.central -side left -anchor nw \
# 		        -fill both -expand 0 -padx 0 -pady 0
# 		if { [winfo exists .gid.central.wins] } {
# 		    pack .gid.central.wins -expand 1 -fill both -side left
# 		} else {
# 		    pack .gid.central.s -fill both -expand yes -side left
# 		}
              GridInside $geomname $w .gid.left "ns"
              set orient vertical
	    }
	    INSIDEBOTTOM {
# 		pack $w -in .gid.central -side bottom -anchor sw \
# 		        -fill both -expand 0 -padx 0 -pady 0
# 		if { [winfo exists .gid.central.wins] } {
# 		    pack .gid.central.wins -expand 1 -fill both -side bottom
# 		} else {
# 		    pack .gid.central.s -fill both -expand yes -side bottom
# 		}
              GridInside $geomname $w .gid.bottom "we"
              set orient horizontal
	    }
	    INSIDERIGHT {
# 		pack $w -in .gid.central -side right -anchor ne \
# 		        -fill both -expand 0 -padx 0 -pady 0
# 		if { [winfo exists .gid.central.wins] } {
# 		    pack .gid.central.wins -expand 1 -fill both -side right
# 		} else {
# 		    pack .gid.central.s -fill both -expand yes -side right
# 		}
              GridInside $geomname $w .gid.right "ns"
		set orient vertical
	    }
	}
	if { $widget != ""} {
	    pack configure $w -before $widget
	}
    }
#     bind $w <Configure> [list SaveNewBitmapsGeom $name \
#             $w toolbarmacros::Create PrePost]
    if { [winfo class $w] == "Toplevel" } {
	bind $w <Leave> [list wm title $w $title]
	foreach i [bind .gid] {
	    bind $w $i [bind .gid $i]
	}
    }

    #toolbar_create $w
    bind $w <Configure> [list +toolbarmacros::toolbar_create $w]

    
    GidHelpT $w $title
    return $w
}

proc toolbarmacros::toolbar_create { w } {
    variable afterid

    if { [info exists afterid] } { after cancel $afterid }
    set afterid [after idle toolbarmacros::_toolbar_createDo $w]
}

proc toolbarmacros::_toolbar_createDo { w { force "" } } {
    global GidPriv
    variable images
    variable afterid
    variable toolbar_last_len
    variable preferences

    unset -nocomplain afterid
    update idletasks

    if { ![winfo exists $w] } { return }

    if { [winfo width $w] > [winfo height $w] } {
	set orient horizontal
    } else { set orient vertical }

    switch $orient {
	vertical {
	    set len [winfo height $w]
	}
	horizontal {
	    set len [winfo width $w]
	}
    }
    if { $force eq "" && [winfo exists $w.w0] && [info exists toolbar_last_len] && \
	$toolbar_last_len == $len } {
	return
    }
    set toolbar_last_len $len

    switch $orient {
	vertical {
	    foreach "sticky_s orient_s padx_s pady_s" [list new horizontal 0 2] break
	}
	horizontal {
	    foreach "sticky_s orient_s padx_s pady_s" [list wns vertical 2 0] break
	    #Separator $w.sep1 -orient vertical
	}
    }

    set cmds [list $images(record-22) [_ "Record macro"] \
	    [list toolbarmacros::StartRecording $w toolbar] ""\
	    $images(edit-22) [_ "Edit macros..."] toolbarmacros::MacrosWindow "" \
	    - - - -]

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    set prepost [lindex [GiD_Info Project] 4]
    set lst ""
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name
	if { !$macrosdata($name,InToolbar) || !$macrosdata($name,Active) } {
	    continue
	}
	if { $prepost eq "GEOMETRYUSE" && $macrosdata($name,PrePost) eq "post" } {
	    continue
	}
	if { $prepost eq "POSTUSE" && $macrosdata($name,PrePost) eq "pre" } {
	    continue
	}
	lappend lst [list $macrosdata($i) $name]
    }
    set lst [lsort -integer -index 0 $lst]
#     for { set i [llength $lst] } { $i < 8 } { incr i } {
#         lappend lst [list $i -]
#     }
    set idximage 0
    foreach i $lst {
	set name [lindex $i 1]
	if { $name eq "-" } {
	    if { $idximage < 8 } { incr idximage }
	    lappend cmds $images(macro$idximage-22) macro "" ""
	} else {
	    if { $macrosdata($name,Icon) eq "" } {
		if { $idximage < 8 } { incr idximage }
		set image $images(macro$idximage-22)
	    } else { set image [icon_chooser::giveimage $macrosdata($name,Icon)] }
	    set help $name\n[_ $macrosdata($name,Description)]\n
	    append help [_ "Creation date: %s" $macrosdata($name,CreationDate)]\n
	    append help [_ "Modification date: %s" $macrosdata($name,ModificationDate)]\n
	    lappend cmds $image [_ $macrosdata($name,Description)] \
		[list ::toolbarmacros::macrospace::$name] $name
	}
    }
    set inum 0
    if { [winfo exists $w.w0] } {
       while { [winfo exists $w.w$inum] } {
	    if { $force ne "" } {
		destroy $w.w$inum
	    } else {
		grid forget $w.w$inum
	    }
	    incr inum
	}
    }


    set maxrow 0
    set maxcol 0
    set rlen 0
    foreach "row col inum" [list 0 0 0] break
    foreach "img help cmd macroname" $cmds {        
	if { $img eq "-" } {
	    if { ![winfo exists $w.w$inum] } {
		Separator $w.w$inum -orient $orient_s
	    }
	    foreach "sticky padx pady" [list $sticky_s $padx_s $pady_s] break
	    switch $orient horizontal { set isseparator(col,$col) 1 } vertical { set isseparator(row,$row) 1 }
	    
	} else {
	    if { ![winfo exists $w.w$inum] } {
		button $w.w$inum -image $img -bg $GidPriv(Color,ToolbarBg) \
		    -relief flat -bd 0 -highlightthickness 0 \
		    -command $cmd -width 24 -height 24 
		    #-overrelief raised
		ChangeToNewLook $w.w$inum
		GidHelp $w.w$inum $help
		if { $macroname ne "" } {
		    bind $w.w$inum <3> [namespace code [list ToolbarContextual \
		                %W $macroname $help %X %Y]]
		}
	    }
	    foreach "sticky padx pady" [list nw 0 0] break
	}
	switch $orient {    
	    vertical {
		set len [winfo height $w]
		#set rlen [winfo reqheight $w]
		incr rlen [winfo reqheight $w.w$inum]
	    }        
	    horizontal {
		set len [winfo width $w]
		#set rlen [winfo reqwidth $w]
		incr rlen [winfo reqwidth $w.w$inum]
	    }            
	}
	if { $inum > 1 && $len < $rlen } { 
	    switch $orient {
		vertical {
		    if { $preferences(toolbar_one_col) } {
		        break
		    }
		    incr col
		    set row 0
		    set rlen [winfo reqheight $w.w$inum]
		}
		horizontal {
		    if { $preferences(toolbar_one_col) } {
		        break
		    } 
		    incr row
		    set col 0
		    set rlen [winfo reqwidth $w.w$inum]
		}                
	    }            
	}

	switch $orient {
	    vertical { 
		if { $maxrow < $row} { set maxrow $row }
	    }
	    horizontal { 
		if { $maxcol < $col} { set maxcol $col }
	    }
	}        

	grid $w.w$inum -row $row -column $col -sticky $sticky \
	    -padx $padx -pady $pady
	grid rowconfigure $w $row -weight 0
	grid columnconfigure $w $col -weight 0       
		
	switch $orient vertical { incr row } horizontal { incr col }
     
	if { ([info exists isseparator(row,$row)] || [info exists isseparator(col,$col)]) && $img != "-" } {
	    switch $orient vertical { incr row } horizontal { incr col }
	}        
	incr inum
    }
    grid rowconfigure $w $maxrow -weight 1
    grid columnconfigure $w $maxcol -weight 1

#     for { set i [expr {$inum-1}] } { $i >= 2 } { incr i -1 } {
#         update idletasks
#         switch $orient {
#             vertical {
#                 set len [winfo height $w]
#                 set rlen [winfo reqheight $w]
#             }
#             horizontal {
#                 set len [winfo width $w]
#                 set rlen [winfo reqwidth $w]
#             }
#         }
#         if { $len < $rlen } { grid forget $w.w$i } else { break }
#     }
}

proc toolbarmacros::ToggleDetailsFrame { w { changegeometry 1 } } {

    wm withdraw $w
    update idletasks

    if { ![winfo exists $w.pw] } {
	set wf1 [winfo width $w.f1]
	if { $wf1 <= 1 } { set wf1 [winfo reqwidth $w.f1] }
	grid forget $w.f1
#         panedwindow $w.pw
#         $w.pw add $w.f1 -sticky nsew -padx 1 -pady 1 -width $wf1
#         $w.pw add $w.f2 -sticky nsew -padx 1 -pady 1

	PanedWindow $w.pw
	set p1 [$w.pw add -weight $wf1 -minsize 70]
	set p2 [$w.pw add -weight [winfo reqwidth $w.f2] -minsize 190]
	grid $w.f1 -in $p1 -sticky nsew -padx 3 -pady 3
	grid columnconfigure $p1 0 -weight 1
	grid rowconfigure $p1 0 -weight 1
	grid $w.f2 -in $p2 -sticky nsew -padx 3 -pady 3
	grid columnconfigure $p2 0 -weight 1
	grid rowconfigure $p2 0 -weight 1

	raise $w.f1
	raise $w.f2
	grid $w.pw -sticky nsew -row 0 -column 0 -padx 0 -pady 0
	set newwidth [expr {[winfo width $w]+[winfo reqwidth $w.f2]+0}]
	InitWindow_ModifyInitComm $w TMacrosWindowGeom \
	    [list toolbarmacros::MacrosWindow complete]
    } else {
	destroy $w.pw
	grid $w.f1 -sticky nsew -row 0 -column 0 -padx 3 -pady 3
	set newwidth [expr {[winfo width $w.f1]}]
	InitWindow_ModifyInitComm $w TMacrosWindowGeom \
	    [list toolbarmacros::MacrosWindow simple]
    }

    if { $changegeometry } { wm geometry $w ${newwidth}x[winfo height $w] }
    wm deiconify $w
}

proc toolbarmacros::ToolbarContextual { button name help x y } {
    
   set menu $button.menu
    destroy $menu
    menu $menu -tearoff 0
    
    $menu add command -label [_ "Help"] -command \
	    [list DisplayGidHelp $button $help menu $x $y]
    $menu add command -label [_ "Configure toolbars"] -command \
	TOBegin
    $menu add separator
    $menu add command -label [_ "Rename"] -command \
	[namespace code [list _rename_copy_macro .gid.tmacros rename $name]]
    $menu add command -label [_ "Change icon"] -command \
	[namespace code [list ChangeToolbarIcon $button $name]]
    $menu add command -label [_ "Add accelerator"] -command \
	[namespace code [list FastAcceleratorsWindow $button $name]]
    $menu add command -label [_ "Delete"] -command \
	[namespace code [list DeleteMacro $button [list $name]]]
    $menu add separator
    $menu add command -label [_ "Edit window"] -command \
	[namespace code [list MacrosWindow undefined [list $name]]]

    tk_popup $menu $x $y
}

proc toolbarmacros::StartRecording { w type args } {
    variable images
    if { $type eq "toolbar" } {
	foreach i [winfo children $w] {
	    if { [winfo class $i] eq "Button" && \
		[$i cget -image] eq $images(record-22) } {
		$i configure -image $images(recordend-22) -command \
		    [list toolbarmacros::EndRecording $w $type]
		GidHelp $i [_ "Press here to end recording macro"]
		break
	    }
	}
    } elseif { $type eq "window" } {
	foreach "bbox idx" $args break
	$bbox itemconfigure $idx -image $images(recordend-16) -command \
	    "[list toolbarmacros::EndRecording $w $type] $args" \
	    -helptext [_ "Press here to end recording macro"]

    }
    .central.s recordbatch begin
}

proc toolbarmacros::_ReorderMacrosNumbers { { hole_interval "" } } {
    
    if { $hole_interval eq "" } {
	foreach "h_start h_end" [list 0 0] break
    } else {
	foreach "h_start h_end" $hole_interval break
    }

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    set lst ""
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name
	lappend lst [list $macrosdata($i) $name]
    }
    set lst [lsort -integer -index 0 $lst]
    set idx 1
    for { set i 0 } { $i < [llength $lst] } { incr i } {
	while { $idx >= $h_start && $idx <= $h_end } { incr idx }
	lset lst $i 0 $idx
	incr idx
    }
    foreach i $lst {
	foreach "number name" $i break
	set macrosdata($name,Number) $number
    }
}

proc toolbarmacros::EndRecording { w type args } {
    variable images
    if { $type eq "toolbar" } {
	foreach i [winfo children $w] {
	    if { [winfo class $i] eq "Button" && \
		[$i cget -image] eq $images(recordend-22) } {
		$i configure -image $images(record-22) -command \
		    [list toolbarmacros::StartRecording $w $type]
		GidHelp $i [_ "Record macro"]
		break
	    }
	}
    } elseif { $type eq "window" } {
	foreach "bbox idx" $args break
	$bbox itemconfigure $idx -image $images(record-16) -command \
	    "[list toolbarmacros::StartRecording $w $type] $args" \
	    -helptext [_ "Record macro"]

    }
    set data [.central.s recordbatch end]

    regsub -all {(?n)\*{5}SAVE MARK.*$} $data {} data
    regsub -all {\*{5}SAVE STATE.*?\*{5}END SAVE STATE} $data {} data
    CreateMacro $data
}

proc toolbarmacros::CreateMacro { data { name "" } { desc "" } { acc "" } { update yes } } {

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    _ReorderMacrosNumbers [list 1 1]
    
    set macrosnames ""
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name_in
	lappend macrosnames $name_in
    }
    if { $name eq "" } {
	set i 1
	while 1 {
	    if { [lsearch -exact $macrosnames Macro$i] == -1 } { break }
	    incr i
	}
	set name Macro$i
    }

    if { $desc eq "" } {
	regsub -all {\mescape\M} $data {} desc
	regsub -all {[ \t]+} $desc { } desc
	set desc [string trim $desc]
	if { [string length $desc] > 200 } {
	    set desc [string range $desc 0 196]...
	}
    }
    set now [clock format [clock seconds] -format "%Y-%m-%d %H:%M:%S"]

    set macrosdata($name,Number) 1
    set macrosdata($name,Accelerators) $acc
    set macrosdata($name,Description) $desc
    set macrosdata($name,Group) ""
    set macrosdata($name,InToolbar) 1
    set macrosdata($name,Active) 1
    set macrosdata($name,PrePost) prepost
    set macrosdata($name,Icon) ""
    set macrosdata($name,CreationDate) $now
    set macrosdata($name,ModificationDate) $now
    set macrosdata($name,IsStandard) 0

    set body ""
    set body_pr ""
    foreach line [split $data \n] {
	if { [string trim $line] ne "" } {
	    if { [string match -np-* $line] } {
		if { $body_pr ne "" } {
		    append body "GiD_Process $body_pr\n"
		    set body_pr ""
		}
		append body "[string trim [string range $line 4 end]]\n"
	    } else {
		append body_pr "$line "
		
	    }
	}
    }
    if { $body_pr ne "" } {
	append body "GiD_Process $body_pr\n"
    }
    set macrosdata($name,Code) $body
    namespace eval macrospace [list proc $name {} $body]

    if { $update eq "yes" } {
	_update_MacrosList .gid.tmacros 0
	_update_toolbar
	SaveMacros
    }
}

proc toolbarmacros::ApplyMacrosBindings {} {
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    set prepost [lindex [GiD_Info Project] 4]
    set lst ""
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name
	lappend lst [list $macrosdata($i) $name]
    }
    foreach i [lsort -decreasing -integer -index 0 $lst] {
	set name [lindex $i 1]
	if { !$macrosdata($name,Active) } { continue }
       if { $prepost eq "GEOMETRYUSE" && $macrosdata($name,PrePost) eq "post" } {
	    continue
	}
	if { $prepost eq "POSTUSE" && $macrosdata($name,PrePost) eq "pre" } {
	    continue
	}
	foreach j $macrosdata($name,Accelerators) {
	    set bind [_manage_accelerators_short2long $j]
	    bind .gid $bind [list ::toolbarmacros::macrospace::$name]
	}
    }
}

proc toolbarmacros::FastAcceleratorsWindow { button name } {
    variable images

    set w [toplevel $button.faw]
    wm overrideredirect $w 1
    set x [expr {[winfo rootx $button]+[winfo width $button]}]
    set y [winfo rooty $button]
    wm geometry $w +$x+$y
    wm title $w [_ "Enter accelerator"]

    $w configure -bd 2 -relief groove

    button $w.l1 -image [icon_chooser::giveimage stop-16] -command [list destroy $w] \
	-relief flat -overrelief raised -bd 1

    label $w.l2 -text [_ "Accelerators:"]
    label $w.l22 -textvariable toolbarmacros::macrospace::macrosdata($name,Accelerators)
    label $w.l3 -text [_ "New accelerator:"]
    entry $w.e1 -width 7
    bind $w.e1 <KeyPress> "[namespace code [list _enter_accelerators_here_do \
		$w.e1 %K %A]] ; break"
    bind $w.e1 <Return> [namespace code [list _manage_accelerators_add $w.e1 $name]]
    $w.e1 del 0 end
    ButtonBox $w.b1 -spacing 0 -padx 1 -pady 1 -homogeneous 1
    $w.b1 add -image $images(ok-16) -helptext [_ "Define new accelerator"] \
	-relief link -borderwidth 1 -padx 1 -pady 1 -command \
	[namespace code [list _manage_accelerators_add $w.e1 $name]]

    grid $w.l1 - - -sticky ne
    grid $w.l2 $w.l22 - -pady "5 2" -sticky nw -padx 2
    grid $w.l3 $w.e1 $w.b1 -sticky nw -padx 2 -pady "0 3"

    focus $w.e1
    grab $w
    bind $w <Escape> [list destroy $w]
    bind $w.e1 <Escape> "[list destroy $w] ; break"
}

proc toolbarmacros::_update_toolbar {} {
    _toolbar_createDo .gid.bitmapsMacrosToolbar force
}

proc toolbarmacros::_update_MacrosList { w { selnums "" } { selnames "" } } {
    variable tablelists
    variable filterentry
    variable filterentry_onlygroup
    variable afterfilterid
    variable images
    variable groups

    if { [info exists afterfilterid] } { unset afterfilterid }
    if { ![winfo exists $w] } { return }
    set t $tablelists($w)

    if { $selnums eq "" } {
	set yviewargs [list moveto [lindex [$t yview] 0]]
    } else {
	# to make treectnrl work
	set yviewargs [list moveto 0]
	#set yviewargs [lindex $selnums 0]
    }
    if { $selnums eq "" && $selnames eq "" } {
	set selnums [$t curselection]
    }
    if { $selnums ne "" } {
	foreach i $selnums {
	    lappend selnames [lindex [$t get $i] 1]
	}
    }

    $t delete 0 end
   
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    set lst ""
    set groups ""
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name
	lappend lst [list $macrosdata($i) $name]
	if { $macrosdata($name,Group) ne "" && \
	    [lsearch -exact $groups $macrosdata($name,Group)] == -1 } {
	    lappend groups $macrosdata($name,Group)
	}
    }
    set lst [lsort -integer -index 0 $lst]

    if { [catch { regexp $filterentry {} }] } {
	set rex (?qi)$filterentry
    } else { set rex (?i)$filterentry }
    foreach i $lst {
	foreach "number name" $i break
	if { $filterentry_onlygroup == 0 && $filterentry ne "" && ![regexp $rex $name] && \
	    ![regexp $rex $macrosdata($name,Description)] && \
	    ![regexp $rex $macrosdata($name,Group)] } {
	    continue
	}
	if { $filterentry_onlygroup == 1 && ![regexp $rex $macrosdata($name,Group)] } {
	    continue
	}

	if { $macrosdata($name,Icon) ne "" } {
	    set icon [icon_chooser::giveimage $macrosdata($name,Icon)]
	} else { set icon $images(macro1-22) }
	set intimg [expr {$macrosdata($name,InToolbar)?$images(ok-16):$images(transparent-16)}]
	set actimg [expr {$macrosdata($name,Active)?$images(ok-16):$images(transparent-16)}]
	set ppimg $images($macrosdata($name,PrePost)-24)
	$t insert end [list $number [list $icon $name] $macrosdata($name,Accelerators) \
		[list $intimg $macrosdata($name,InToolbar)] \
		[list $actimg $macrosdata($name,Active)] \
		[list $ppimg $macrosdata($name,PrePost)] \
		$macrosdata($name,Group) \
		[_ $macrosdata($name,Description)] \
		$macrosdata($name,CreationDate) \
		$macrosdata($name,ModificationDate)]


#         $t insert end [list $number $name $macrosdata($name,Accelerators) \
#                 $macrosdata($name,InToolbar) $macrosdata($name,Active) \
#                 "" \
#                 $macrosdata($name,Group) \
#                 $macrosdata($name,Description) \
#                 $macrosdata($name,CreationDate) $macrosdata($name,ModificationDate)]
#         set row [expr {[$t index end]-1}]
#         if { $macrosdata($name,InToolbar) } {
#             $t cellconfigure $row,3 -image $images(ok-16)
#         } else {
#             $t cellconfigure $row,3 -image $images(transparent-16)
#         }
#         if { $macrosdata($name,Active) } {
#             $t cellconfigure $row,4 -image $images(ok-16)
#         } else {
#             $t cellconfigure $row,4 -image $images(transparent-16)
#         }
#         if { $macrosdata($name,Icon) ne "" } {
#             $t cellconfigure $row,5 -image \
#                 [icon_chooser::giveimage $macrosdata($name,Icon)]
#         }
	if { [lsearch -exact $selnames $name] != -1 } {
	    $t selection_r add last
	}
    }
   set col [$t sortcolumn]
   if { $col != -1 } {
       $t sortbycolumn $col -[$t sortorder]
    }
    if { [$t numitems] > 1 } {
	foreach i $selnums {        
	    $t selection_r add $i        
	}
    } else {
	$t selection_r clear
    }
    eval [list $t yview] $yviewargs

    _changeMacrosList_selection $w
    ApplyMacrosBindings
}

proc toolbarmacros::_filter_MacrosList { w } {
    variable afterfilterid
    if { [info exists afterfilterid] } {
	after cancel $afterfilterid
    }
    set afterfilterid [after 100 [namespace code \
		[list _update_MacrosList $w ""]]]
}

proc toolbarmacros::_changeMacrosList_selection_handler { w { update_details yes } } {
    variable afterCMLid

    if { [info exists afterCMLid] } {
	after cancel $afterCMLid
    }
    set afterCMLid [after 100 [namespace code \
		[list _changeMacrosList_selection $w $update_details]]]
}
proc toolbarmacros::_changeMacrosList_selection { w { update_details yes } } {
    variable tablelists
    variable macroname
    variable macroaccelerators
    variable macrointoolbar
    variable macroactive
    variable macroprepost
    variable macrodescription
    variable macrogroup
    variable macrocode
    variable iconbuttons
    variable images
    variable Dragging

    set t $tablelists($w)

    set id [lindex [$t curselection] 0]
    if { [info exists Dragging] || [llength [$t curselection]] != 1 } {
	if { $update_details } { update_macrodetails $w }
	set macroaccelerators ""
	set macrointoolbar 0
	set macroactive 0
	set macroprepost prepost
	set macrodescription ""
	set macrogroup ""
	set macrocode ""
	# macroname must be at the end as it activates a handler
	set macroname ""
	return
    }
    set newmacroname [lindex [$t get $id] 1]
    if { $macroname eq $newmacroname && $update_details eq "yes" } { return }
    if { $update_details } { update_macrodetails $w }

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    set lst ""
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name
	lappend lst [list $macrosdata($i) $name]
    }
    set macronames [lsort -integer -index 0 $lst]
    set macroname $newmacroname
    set macroaccelerators $macrosdata($macroname,Accelerators)
    set macrointoolbar $macrosdata($macroname,InToolbar)
    set macroactive $macrosdata($macroname,Active)
    set macroprepost $macrosdata($macroname,PrePost)
    set macrogroup $macrosdata($macroname,Group)
    set macrodescription $macrosdata($macroname,Description)
    set macrocode $macrosdata($macroname,Code)
    #set macrocode [string trim [namespace eval macrospace [list info body $macroname]]]

    set icon $macrosdata($macroname,Icon)
    if { $icon ne "" } {
	$iconbuttons($w) configure -image [icon_chooser::giveimage $icon] \
	    -helptext $icon\n[_ "Press to choose icon"]
    } else {
	$iconbuttons($w) configure -image $images(macro1-22) \
	    -helptext [_ "Press to choose icon"]
    }
}

proc toolbarmacros::update_macrodetails { w } {
    variable macroname
    variable macroaccelerators
    variable macrointoolbar
    variable macroactive
    variable macroprepost
    variable macrogroup
    variable macrodescription
    variable macrocode
    variable iconbuttons
    variable images

    if { ![winfo exists $w] } { return }

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    if { ![info exists macroname] || $macroname eq "" || \
	![info exists macrosdata($macroname,Number)] } {
	return
    }

    set changes 0
    if { $macrosdata($macroname,Accelerators) ne $macroaccelerators } {
	foreach j $macrosdata($macroname,Accelerators) {
	    set bind [_manage_accelerators_short2long $j]
	    bind .gid $bind ""
	}
	set macrosdata($macroname,Accelerators) $macroaccelerators
	set changes 1
    }
    if { $macrosdata($macroname,InToolbar) != $macrointoolbar } {
	set macrosdata($macroname,InToolbar) $macrointoolbar
	set changes 1
    }
    if { $macrosdata($macroname,Active) != $macroactive } {
	set macrosdata($macroname,Active) $macroactive
	set changes 1
	if { !$macroactive } {
	    foreach j $macrosdata($macroname,Accelerators) {
		set bind [_manage_accelerators_short2long $j]
		bind .gid $bind ""
	    }
	}
    }
    if { $macrosdata($macroname,PrePost) ne $macroprepost } {
	set macrosdata($macroname,PrePost) $macroprepost
	set changes 1
    }
    if { $macrosdata($macroname,Description) ne $macrodescription } {
	set macrosdata($macroname,Description) $macrodescription
	set changes 1
    }
    if { $macrosdata($macroname,Group) ne $macrogroup } {
	set macrosdata($macroname,Group) $macrogroup
	set changes 1
    }
    #set macrocode_old [string trim [namespace eval macrospace [list info body $macroname]]]
    if { $macrosdata($macroname,Code) ne $macrocode } {
	set macrosdata($macroname,Code) $macrocode
	namespace eval macrospace [list proc $macroname "" $macrocode]
	set changes 1
    }

    if { [$iconbuttons($w) cget -image] eq $images(macro1-22) } {
	set icon ""
    } else {
	set icon [lindex [split [$iconbuttons($w) cget -helptext] \n] 0]
    }
    if { $icon ne $macrosdata($macroname,Icon) } {
	set macrosdata($macroname,Icon) $icon
	set changes 1
    }

    if { $changes } {
	set now [clock format [clock seconds] -format "%Y-%m-%d %H:%M:%S"]
	set macrosdata($macroname,ModificationDate) $now
	_update_MacrosList $w
	_update_toolbar
	SaveMacros
	ApplyMacrosBindings
    }
}


proc toolbarmacros::_rename_copy_macro { w what { name "" } } {
    variable macroname

    if { $name eq "" } { set name $macroname }
    update_macrodetails $w

    switch $what {
	rename { set tt [_ "Enter new macro name"] }
	copy { set tt [_ "Enter name for the copied macro"] }
    }

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    set lst ""
    set maxnumber 0
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name_in
	lappend lst $name_in
	if { $macrosdata($i) > $maxnumber } { set maxnumber $macrosdata($i) }
    }
    if { [winfo exists $w] } {
	set win $w._ask
    } else { set win .gid._ask }
    dialogwin_snit $win -title $tt -entrytype entry \
	-entrylabel [_ "Name:"] -entrytext $tt: -entrydefault $name
    set action [$win createwindow]
    while 1 {
	if { $action <= 0 } { 
	    destroy $win
	    return
	}
	set newname [string trim [$win giveentryvalue]]
	if { $newname eq "" } {
	    WarnWin [_ "Enter a name"] $w
	} elseif { [string match *,* $newname] } {
	    WarnWin [_ "Character ',' is not allowed in name"] $w
	} elseif { [lsearch -exact $lst $newname] != -1 } {
	    WarnWin [_ "Macro name '%s' already exists" $newname] $w
	} else { break }
	set action [$win waitforwindow]
    }
    destroy $win

    set now [clock format [clock seconds] -format "%Y-%m-%d %H:%M:%S"]
    if { $what eq "rename" } {
	foreach j $macrosdata($name,Accelerators) {
	    set bind [_manage_accelerators_short2long $j]
	    bind .gid $bind ""
	}
	foreach i [list Number Accelerators InToolbar Active PrePost \
		Description Group Icon \
		CreationDate ModificationDate Code IsStandard] {
	    if { [info exists macrosdata($name,$i)] } {
		set macrosdata($newname,$i) $macrosdata($name,$i)
		unset macrosdata($name,$i)
	    }
	}

	namespace eval macrospace [list rename $name $newname]
	set macroname ""
    } else {
	set macrosdata($newname,Number) [incr maxnumber]
	foreach i [list Accelerators InToolbar Active PrePost Group Description \
		Icon Code] {
	    if { [info exists macrosdata($name,$i)] } {
		set macrosdata($newname,$i) $macrosdata($name,$i)
	    }
	}
	set macrosdata($newname,IsStandard) 0        
	namespace eval macrospace [list proc $newname {} \
		$macrosdata($newname,Code)]
	set macrosdata($newname,CreationDate) $now
    }
    set macrosdata($newname,ModificationDate) $now
    _update_MacrosList $w
    _update_toolbar
    SaveMacros
}

proc toolbarmacros::FileOperations {w buttonbox } {

    set menu $w.menu
    destroy $menu
    menu $menu -tearoff 0
    $menu add command -label [_ "Import macros"]... -command \
	[namespace code [list FileOperations_do $w import]]
    $menu add command -label [_ "Import Default macros"] -command \
	[namespace code [list ReadMacrosFromFileAndUpdate [file join $::GIDDEFAULTTCL toolbarmacros_defaults]]]
    $menu add command -label [_ "Export macros"]... -command \
	[namespace code [list FileOperations_do $w export]]
    $menu add command -label [_ "Export selected macros"]... -command \
	[namespace code [list FileOperations_do $w export_selection]]

    update idletasks
    set x [expr {[winfo rootx $buttonbox]+[winfo width $buttonbox]-\
	    [winfo reqwidth $menu]}]
    set y [expr {[winfo rooty $buttonbox]+[winfo height $buttonbox]}]
    tk_popup $menu $x $y
}

proc toolbarmacros::ReadMacrosFromFileAndUpdate { file } {
    if { $file == "" } { return }
    set err [catch { ReadMacrosFromFile $file 0 } errstring]
    if { $err } {
	WarnWin [_ "Could not import macros file '%s' (%s)" $file $errstring]
	return
    }
    icon_chooser::AfterReadMacros
    _update_MacrosList .gid.tmacros
    _update_toolbar
    ApplyMacrosBindings
    toolbarmacros::SaveMacros
}

proc toolbarmacros::FileOperations_do { w what } {
    variable filetoclose
    variable tablelists

    set types [list \
	    [list [_ "TCL files"] [list .tcl]] \
	    [list [_ "All files"] *]]

    if { ![info exists ::DefaultSearchDirectory] || \
	![file isdirectory $::DefaultSearchDirectory]} {
	set ::DefaultSearchDirectory [pwd]
    }
    switch $what {
	import {
	    set file [Browser-ramR file read $w [_ "Import macros"] \
		    "" $types .tcl 0]

	    ReadMacrosFromFileAndUpdate $file
	}
	export {
	    set file [Browser-ramR file write $w [_ "Export macros"] \
		    "" $types .tcl 0]

	    if { $file == "" } { return }
	    set err [catch { SaveMacrosInFile $file } errstring]
	    if { $err } {
		WarnWin [_ "Could not export macros file '%s' (%s)" $file $errstring]
		catch { close $filetoclose }
		return
	    }
	}
	export_selection {
	    set t $tablelists($w)
	    set ids [$t curselection]
	    if { [llength $ids] == 0 } {
		WarnWin [_ "Select one or more macros for exporting them"]
		return
	    }
	    set selnames ""
	    foreach id $ids {
		lappend selnames [lindex [$t get $id] 1]
	    }

	    set file [Browser-ramR file write $w [_ "Export selected macros"] \
		    "" $types .tcl 0]

	    if { $file == "" } { return }
	    set err [catch { SaveMacrosInFile $file $selnames } errstring]
	    if { $err } {
		WarnWin [_ "Could not export macros file '%s' (%s)" $file $errstring]
		catch { close $filetoclose }
		return
	    }
	}
    }
    set ::DefaultSearchDirectory [file dirname $file]
}

proc toolbarmacros::SaveMacros {} {
    variable filetoclose

    set file [file join [file dirname [gidgetdefaultsfile]] Macros.tcl]
    set err [catch { SaveMacrosInFile $file } errstring]
    if { $err } {
	WarnWin [_ "Could not save macros file '%s' (%s)" $file $errstring]
	catch { close $filetoclose }
    }
}

proc toolbarmacros::SaveMacrosInFile { file { names "" } } {
    variable filetoclose
    variable preferences

    set fout [open $file w]
    set filetoclose $fout
		
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    set lst ""
    foreach i [array names macrosdata *,Number] {
	regexp {^[^,]*} $i name
	lappend lst [list $macrosdata($i) $name]
    }
    set icondescs ""
    puts $fout "# Macros file for GiD v1.0\n"
    puts $fout "# Created by GiD version [GiD_Info GiDVersion]\n"
    foreach i [lsort -integer -index 0 $lst] {
	foreach "number name" $i break
	if { [llength $names] } {
	    if { [lsearch -exact $names $name] == -1 } { continue }
	    if { $macrosdata($name,Icon) ne "" } {
		lappend icondescs $macrosdata($name,Icon)
	    }
	}
	if { ![info exists macrosdata($name,IsStandard)] } {
	    set macrosdata($name,IsStandard) 0
	}
	puts $fout "#\[_ \"[regsub -all \n $macrosdata($name,Description) {\n}]\"\]" ;#for translation
	foreach i [list Number Accelerators InToolbar Active PrePost Icon Group \
		Description CreationDate ModificationDate IsStandard] {
	    puts $fout [list set macrosdata($name,$i) $macrosdata($name,$i)]
	}
	puts $fout ""
	puts $fout "[list proc $name] {} \{"
	foreach line [split [string trim $macrosdata($name,Code)] \n] {
	    puts $fout "    $line"
	}
	puts $fout "\}\n"
    }
    icon_chooser::SaveImages $fout $icondescs

    puts $fout [list array set preferences [array get preferences]]

    close $fout
    unset -nocomplain filetoclose
}

proc toolbarmacros::ReadMacros { } {
    set file [file join [file dirname [gidgetdefaultsfile]] Macros.tcl]
    if { [file readable $file] } {
	set err [catch { ReadMacrosFromFile $file 1 } errstring]
	if { $err } {
	    WarnWin [_ "Could not open macros file '%s' (%s)" $file $errstring]
	}
    }
    #update user macros with default macros, to add macros of new versions
    set file [file join $::GIDDEFAULTTCL toolbarmacros_defaults]
    set err [catch { ReadMacrosFromFile $file 0 } errstring]
    if { $err } {
	WarnWin [_ "Could not open macros file '%s' (%s)" $file $errstring]
    }
    icon_chooser::AfterReadMacros
}

proc toolbarmacros::ReadMacrosFromFile { file readpreferences } {
    variable macrosdata
    variable preferences

    namespace eval imported_macrospace [list source $file]

    if { $readpreferences } {
	array set preferences [namespace eval imported_macrospace array get preferences]
    }

    foreach i [info commands imported_macrospace::*] {
	lappend imported_commands [namespace tail $i]
    }
    set imported_commands [lsort $imported_commands]

    upvar #0 toolbarmacros::imported_macrospace::macrosdata imported_macrosdata

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata    

    #try to mantain the current macros settings, and after add the new ones
    foreach i [info commands macrospace::*] {
	set name [namespace tail $i]
	if { [lsearch -sorted $imported_commands $name] != -1 } {
	    #set the imported body
	    set body [string trim [namespace eval imported_macrospace [list info body $name]]]
	    regsub -all {\n\s{4}} $body "\n" body
	    if { $macrosdata($name,Code) != $body } {
		set macrosdata($name,Code) $body
	    }
	}
	if { [info exists imported_macrosdata($name,IsStandard)] } {
	    set macrosdata($name,IsStandard) $imported_macrosdata($name,IsStandard)
	    if { $imported_macrosdata($name,IsStandard) == "1" && \
		    $imported_macrosdata($name,Active) == "0" && \
		    $imported_macrosdata($name,InToolbar) == "0" } {
		#is an standard macro that is hidden in the imported version
		set macrosdata($name,Active) 0
		set macrosdata($name,InToolbar) 0                
	    }
	} elseif { [info exists macrosdata($name,IsStandard)] && $macrosdata($name,IsStandard) == "1" } {
	    #is an standard macro that is removed in the imported version
	    set macrosdata($name,Active) 0
	    set macrosdata($name,InToolbar) 0                
	}
    }
    #add the new ones, without repeat an used number
    set maxnumber 0
    foreach i [array names macrosdata *,Number] {
	if { $macrosdata($i) > $maxnumber } { set maxnumber $macrosdata($i) }
	set usednumber($macrosdata($i)) 1
    }
    foreach i [array names imported_macrosdata *,Number] {
	set name [string range $i 0 end-7]
	if { ![info exists macrosdata($name,Number)] } {
	    foreach j [array names imported_macrosdata $name,*] {
		set macrosdata($j) $imported_macrosdata($j)    
	    }
	    #set the imported body
	    set body [string trim [namespace eval imported_macrospace [list info body $name]]]
	    regsub -all {\n\s{4}} $body "\n" body
	    set macrosdata($name,Code) $body            
	    rename imported_macrospace::$name macrospace::$name
	    if { [info exists usednumber($macrosdata($name,Number))] } {
		set macrosdata($name,Number) [incr maxnumber]    
	    } 
	    if { $macrosdata($name,Number) > $maxnumber } {
		set maxnumber $macrosdata($name,Number)
	    }
	    set usednumber($macrosdata($name,Number)) 1 
	}     
    }

    unset imported_macrosdata
    foreach i [info commands imported_macrospace::*] {
	rename $i ""
    }

    #avoid image names with spaces (problems using tile)    
    foreach name [array names ::icon_chooser::ImportedDates] {
	if { [regsub -all { } $name {_} newname] > 0 } {
	    set ::icon_chooser::ImportedDates($newname) $::icon_chooser::ImportedDates($name)
	    array unset ::icon_chooser::ImportedDates($name)
	}  
    }
}

proc toolbarmacros::UpdateAllMacrosForTest { } {
    upvar #0 toolbarmacros::imported_macrospace::macrosdata imported_macrosdata
    unset -nocomplain imported_macrosdata
    foreach i [info commands imported_macrosdata::*] {
	rename $i ""
    }

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata  
    unset -nocomplain macrosdata
    foreach i [info commands macrospace::*] {
	rename $i ""
    }
    toolbarmacros::ReadMacros
}

proc toolbarmacros::TransferMacrosFromOldSystem {} {
    global MacrosList

    foreach i [array names MacrosList] {
	regexp [join [list "ACCELERATORS\n(.*)END ACCELERATORS.*" \
		"COMMENTS\n(.*)END COMMENTS.*ACTION\n(.*)END ACTION"] ""] \
		$MacrosList($i) {} acc comments action
	foreach j $acc {
	    if { [regexp {KeyPress} $j] } {
		regsub -all -- {KeyPress} $j {Key} j
	    } elseif { ![regexp {Key} $j] } {
		regsub -all -- {-\d$} $j {-Key&} j
	    }
	}
	CreateMacro [string trim $action] $i [string trim $comments] [string trim $acc] no
    }
    _update_MacrosList .gid.tmacros
    _update_toolbar
    SaveMacros
}

proc toolbarmacros::PlayMacro { w } {
    variable tablelists

    set t $tablelists($w)
    set id [lindex [$t curselection] 0]
    if { [llength $id] != 1 } {
	WarnWin [_ "Select only one macro for playing it"]
	return
    }
    set name [lindex [$t get $id] 1]
    macrospace::$name
}

proc toolbarmacros::DeleteMacro { w { names "" } } {
    variable tablelists

    if { $names eq "" } {
	set t $tablelists($w)
	foreach id [$t curselection] {
	    lappend names [lindex [$t get $id] 1]
	}
    }
    switch [llength $names] {
	0 {
	    WarnWin [_ "Select one or more macros for deleting them"]
	    return
	}
	1 {
	    set name [lindex $names 0]
	    set tt [_ "Are you sure to delete macro '%s?" $name]
	}
	default {
	    set tt [_ "Are you sure to delete %d macros?" [llength $names]]
	}
    }
    set retval [tk_messageBox -parent $w -message $tt -title [_ "Warning"] \
	    -icon question -default ok -type okcancel]
    if { $retval eq "cancel" } { return }
    
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    set standardmacros ""
    foreach name $names {
	if { $macrosdata($name,IsStandard) == "1" } {
	    set macrosdata($name,Active) 0
	    set macrosdata($name,InToolbar) 0
	    lappend standardmacros $name
	    continue
	}
	foreach j $macrosdata($name,Accelerators) {
	    set bind [_manage_accelerators_short2long $j]
	    bind .gid $bind ""
	}
	foreach i [list Number Accelerators Group Description InToolbar Active \
		PrePost Icon CreationDate ModificationDate] {
	    unset macrosdata($name,$i)
	}
	namespace eval macrospace [list rename $name {}]
    }
    if { $standardmacros != "" } {
	WarnWin [_ "Standard macros can't be deleted. Set %d macros to inactive" [llength $standardmacros]]
    }
    _ReorderMacrosNumbers
    _update_MacrosList .gid.tmacros
    _update_toolbar
    SaveMacros
}

proc toolbarmacros::MoveMacroUpDown { w what } {
    variable tablelists
    variable macroname

     # to activate trace
    set macroname ""

    set t $tablelists($w)
    set ids [$t curselection]
    if { [llength $ids] == 0 } {
	WarnWin [_ "Select one or more macros for moving them"]
	return
    }
    set sellst ""
    set selnames ""
    foreach id $ids {
	lappend sellst [lrange [$t get $id] 0 1]
	lappend selnames [lindex [$t get $id] 1]
    }
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    if { $what eq "top" || $what eq "bottom" || [lindex $what 0] eq "pos" } {
	set newnum [lindex [$t get [lindex $what 1]] 0]
	switch -glob -- $what {
	    top {
		_ReorderMacrosNumbers [list 1 [llength $selnames]]
		set idx 1
	    }
	    bottom {
		_ReorderMacrosNumbers
	    }
	    pos* {
		_ReorderMacrosNumbers [list $newnum \
		        [expr {$newnum+[llength $selnames]-1}]]
		set idx $newnum
	    }
	}
	foreach "lst maxnumber" [list "" 0] break
	foreach i [array names macrosdata *,Number] {
	    regexp {^[^,]*} $i name
	    lappend lst [list $macrosdata($i) $name]
	    if { $macrosdata($i) > $maxnumber } { set maxnumber $macrosdata($i) }
	}
	if { $what eq "bottom" } { set idx [expr {$maxnumber+1}] }

	foreach i [lsort -integer -index 0 $lst] {
	    set name [lindex $i 1]
	    if { [lsearch -exact $selnames $name] != -1 } {
		set macrosdata($name,Number) $idx
		incr idx
	    }
	}
	_ReorderMacrosNumbers
    } else {
	if { $sellst ne [lsort -integer -index 0 $sellst] } {
	    WarnWin [_ "To apply this operation, list must be sorted by number"]
	    return
	}
	
	set maxnumber 0
	foreach i [array names macrosdata *,Number] {
	    regexp {^[^,]*} $i name
	    set number2name($macrosdata($i)) $name
	    if { $macrosdata($i) > $maxnumber } { set maxnumber $macrosdata($i) }
	}
	if { $what eq "down" } {
	    set sellst [lsort -decreasing -integer -index 0 $sellst]
	    set maxmin [expr {$maxnumber+1}]
	} else { set maxmin 0 }
	foreach i $sellst {
	    foreach "number name" $i break
	    switch $what {
		down {
		    set newnumber [expr {$number+1}]
		    if { $newnumber >= $maxmin } { continue }
		}
		up {
		    set newnumber [expr {$number-1}]
		    if { $newnumber <= $maxmin } { continue }
		}
	    }
	    set newname $number2name($newnumber)
	    set macrosdata($name,Number) $newnumber
	    set macrosdata($newname,Number) $number
	    set number2name($newnumber) $name
	    set number2name($number) $newname
	    set maxmin $newnumber
	}
    }
    _update_MacrosList $w "" $selnames
    _update_toolbar
}

proc toolbarmacros::ChangeToolbarIcon { button name } {
    variable macroname

    # to activate trace
    set macroname ""

    set icondesc [icon_chooser::choose_icon $button]
    if { $icondesc eq "-" } { return }
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    if { $icondesc ne $macrosdata($name,Icon) } {
	set macrosdata($name,Icon) $icondesc
	_update_MacrosList .gid.tmacros
	_update_toolbar
	SaveMacros
    }
}

proc toolbarmacros::ChangeIcon { button } {
    variable macroname
    variable images

    set icondesc [icon_chooser::choose_icon $button]
    if { $icondesc eq "" } {
	 $button configure -image $images(macro1-22) \
	    -helptext [_ "Press to choose icon"]
    } elseif { $icondesc ne "-" } {
	$button configure -image [icon_chooser::giveimage $icondesc] \
	    -helptext $icondesc\n[_ "Press to choose icon"]
    }
}

proc toolbarmacros::EditMacro {} {
    variable macroname
    variable macrocode

    set data "\n[list proc ::toolbarmacros::macrospace::$macroname] {} \{\n"
    foreach line [split [string trim $macrocode] \n] {
	append data "    $line\n"
    }
    append data "\}\n"

    if { ![interp exists ramdebugger] } { interp create ramdebugger }
    ramdebugger eval [list set argv [list -noprefs -rgeometry 500x250 -onlytext]]
	
    set err [catch { package require RamDebugger }]
    if { $err } {
	WarnWin [_ "Package RamDebugger is not installed"]
	return
    }
    interp alias ramdebugger EditEndInRamDebugger "" toolbarmacros::EditEndInRamDebugger

    set tdata ""
    while { $data ne "" } {
	if { [string length $data] > 80 } {
	    for { set i 79 } { $i > 10 } { incr i -1 } {
		if { [string index $data $i] eq " " } { break }
	    }
	    if { $i > 10 } {
		append tdata [string range $data 0 $i] "\\\n        "
		set data [string range $data [expr {$i+1}] end]
	    } else {
		append tdata $data
		set data ""
	    }
	} else {
	    append tdata $data
	    set data ""
	}
    }

    ramdebugger eval [list RamDebugger::OpenFileSaveHandler *$macroname* \
	    $tdata EditEndInRamDebugger]
}

proc toolbarmacros::_enter_accelerators_here_error { entry } {

    set types [list Ctrl-s Alt-a F1 Ctrl-F1 "Ctrl-x Ctrl-a"]
    WarnWin  [_ "Error: Accelerators can be of the types:%s" \n\t[join $types \n\t]] $entry
}
proc toolbarmacros::_enter_accelerators_here_do { entry K A } {

    set data [$entry get]

    # some computers are strange
    if { $K eq "L1" } { set K F11 }
    if { $K eq "L2" } { set K F12 }

    regexp {(Ctrl|Alt)?-?(\S*)\s*(Ctrl|Alt)?-?(\S*)} \
	$data {} ctrl(1) l(1) ctrl(2) l(2)
    
    if { $K == "BackSpace" || $K == "Delete" } {
	$entry delete 0 end
	return
    }
    if { $l(2) != "" } { _enter_accelerators_here_error $entry ; return }
    if { $l(1) != "" } { set i 2 } else { set i 1 }
    
    if { [string match Control* $K] } {
	if { $ctrl($i) == "" } {
	    tk::EntryInsert $entry Ctrl-
	} elseif { $ctrl($i) ne "Ctrl" } { _enter_accelerators_here_error $entry }
    } elseif { [string match Alt* $K] } {
	if { $ctrl($i) == "" } {
	    tk::EntryInsert $entry Alt-
	} elseif { $ctrl($i) ne "Alt" }  { _enter_accelerators_here_error $entry }
    } elseif { $A == "" && [string match F* $K] } {
	tk::EntryInsert $entry "$K "
    } elseif { $A != "" } {
	if { $ctrl($i) != "" } {
	    tk::EntryInsert $entry "$K "
	} else { _enter_accelerators_here_error $entry }
    }
}

proc toolbarmacros::_manage_accelerators_short2long { acc } {
    regsub -all {Ctrl} $acc {Control} acc
    regsub {\s+} $acc {><} acc
    set acc <$acc>
    regsub -all {[^-<]+>} $acc {Key-&} acc
    return $acc
}

proc toolbarmacros::_manage_accelerators_delete  { entry } {
    variable macroname
    variable macroaccelerators

    set acc [string trim [$entry get]]
    regsub -all {\s+} $acc { } acc
    if { $acc eq "" } {
	set macroaccelerators ""
	return
    }
    set ipos [lsearch -exact $macroaccelerators $acc]
    if { $ipos == -1 } {
	WarnWin [_ "Accelerator '%s' is not contained in list '%s" \
		$acc $macroaccelerators] $entry
	return
    }
    set macroaccelerators [lreplace $macroaccelerators $ipos $ipos]
    $entry delete 0 end
}

proc toolbarmacros::_manage_accelerators_add { entry { name "" } } {
    variable macroname
    variable macroaccelerators

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata

    if { $name eq "" } {
	set name $macroname
	set accs $macroaccelerators
	set where editwindow
    } else {
	set where fastwindow
	set accs $macrosdata($name,Accelerators)
    }

    set acc [string trim [$entry get]]
    regsub -all {\s+} $acc { } acc
    if { $acc eq "" } {
	WarnWin [_ "Enter an accelerator in the text space"] $entry
	return
    }
    foreach j $accs {
	if { $acc eq $j } { return }
    }
    set foundname ""
    foreach i [array names macrosdata *,Accelerators] {
	regexp {^[^,]*} $i name_in
	if { $name_in eq $name } { continue }
	foreach j $macrosdata($i) {
	    if { $acc eq $j } {
		set foundname $name_in
		break
	    }
	}
    }
    if { $foundname ne "" } {
	set tt [_ "Accelerator '%s' is currently assigned to macro '%s'. Proceed?" \
		$acc $foundname]
	set w [dialogwin_snit $entry.%AUTO% -title [_ Warning] -entrytext $tt]
	set f [$w giveframe]
	set tt [_ "Note: if two macros have the same accelerator, The one with higher order has precedence"]
	label $f.l1 -text $tt -wraplength 200 -justify left
	grid $f.l1 -sticky nw -padx 3 -pady 10
	$w focusok
	set action [$w createwindow]
	destroy $w
	if { $action < 1 } { return }
    } else {
	set cmd [bind .gid [_manage_accelerators_short2long $acc]]
	if { $cmd ne "" && $cmd ne [list ::toolbarmacros::macrospace::$name] } {
	    set tt [_ "Accelerator '%s' is currently assigned to function '%s'. Proceed?" $acc $cmd]
	    set retval [tk_messageBox -parent $entry -message $tt -title [_ "Warning"] \
		    -icon question -default ok -type okcancel]
	    if { $retval eq "cancel" } { return }
	}
    }
    $entry delete 0 end

    switch $where {
	editwindow {
	    lappend macroaccelerators $acc
	}
	fastwindow {
	    lappend macrosdata($name,Accelerators) $acc
	    _update_MacrosList .gid.tmacros
	    _update_toolbar
	    SaveMacros
	    ApplyMacrosBindings
	}
    }
}

proc toolbarmacros::_add_to_group { w names } {
    variable groups

    if { [winfo exists $w] } {
	set win $w._ask
    } else { set win .gid._ask }
    dialogwin_snit $win -title [_ "Add to group"] -entrytype entry \
	-entrylabel [_ "Group:"] -entrytext [_ "Add to group"]: \
	-entryvalues $groups
    set action [$win createwindow]
    while 1 {
	if { $action <= 0 } { 
	    destroy $win
	    return
	}
	set newname [string trim [$win giveentryvalue]]
	if { [string match *,* $newname] } {
	    WarnWin [_ "Character ',' is not allowed in group name"] $w
	} else { break }
	set action [$win waitforwindow]
    }
    destroy $win

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    foreach name $names {
	set macrosdata($name,Group) $newname
    }
    _update_MacrosList .gid.tmacros
    SaveMacros
}

proc toolbarmacros::Activate_deactivate_intoolbar { what names } {
    variable macroname
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    set macroname ""

    update_macrodetails .gid.tmacros
    foreach name $names {
	switch $what {
	    activate { set macrosdata($name,Active) 1 }
	    deactivate { set macrosdata($name,Active) 0 }
	    intoolbar { set macrosdata($name,InToolbar) 1 }
	    no_intoolbar { set macrosdata($name,InToolbar) 0 }
	    pre { set macrosdata($name,PrePost) pre }
	    post { set macrosdata($name,PrePost) post }
	    prepost { set macrosdata($name,PrePost) prepost }
	}
    }
    _update_MacrosList .gid.tmacros
    _update_toolbar
    SaveMacros
    #ApplyMacrosBindings
}

proc toolbarmacros::EditEndInRamDebugger { name data } {
    variable macroname
    variable macrocode

    set name [string trim $name *]

    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    if { ![info exists macrosdata($name,Number)] } {
	WarnWin [_ "Macro '%s' does not exist" $name]
	return
    }
    set err [catch {
	    namespace eval macrospace $data
	    set body [namespace eval macrospace [list info body $name]]
	} errstring]
    if { $err } {
	WarnWin [_ "There are errors in the code. Correct them and save again (%s)" \
		$errstring]
	return
    }
    regsub -all {\n\s{4}} [string trim $body] "\n" body
    set macrosdata($name,Code) $body

    if { $name eq $macroname } {
	set macrocode $body
    }
}

proc toolbarmacros::ContextualMacrosList { tbl x y X Y} {

#     set y [expr {$y + [winfo y [$tbl bodypath]]}]
#     set id [$tbl nearest $y]
#     if { ![$tbl selection includes $id] } {
#         $tbl selection clear 0 end
#         $tbl selection set $id
#     }

    if { ![$tbl selection_r includes "nearest $x $y"] } {
	$tbl selection_r clear 0 last
	$tbl selection_r add "nearest $x $y"
    }

    set w [winfo toplevel $tbl]
    set menu $tbl.menu
    destroy $menu
    menu $menu -tearoff 0
    
    if { [llength [$tbl curselection]] == 1 } {
	$menu add command -label [_ "Run"] -command \
	    [namespace code [list PlayMacro $w]]
	set name [lindex [$tbl get [$tbl curselection]] 1]
	$menu add command -label [_ "Change icon"] -command \
	    [namespace code [list ChangeToolbarIcon $w $name]]
	$menu add command -label [_ "Rename"] -command \
	    [namespace code [list _rename_copy_macro .gid.tmacros rename $name]]
	$menu add command -label [_ "Copy"] -command \
	    [namespace code [list _rename_copy_macro .gid.tmacros copy $name]]
	$menu add separator
    }
    set names ""
    foreach i [$tbl curselection] {
	lappend names [lindex [$tbl get $i] 1]
    }
    $menu add command -label [_ "Delete"] -command \
	[namespace code [list DeleteMacro $w]]
    $menu add separator
    $menu add command -label [_ "Add to group"] -command \
	[namespace code [list _add_to_group $w $names]]
    $menu add command -label [_ "In toolbar"] -command \
	[namespace code [list Activate_deactivate_intoolbar intoolbar $names]]
    $menu add command -label [_ "Out of toolbar"] -command \
	[namespace code [list Activate_deactivate_intoolbar no_intoolbar $names]]
    $menu add command -label [_ "Activate"] -command \
	[namespace code [list Activate_deactivate_intoolbar activate $names]]
    $menu add command -label [_ "Deactivate"] -command \
	[namespace code [list Activate_deactivate_intoolbar deactivate $names]]

    $menu add cascade -label [_ "PrePost"] -menu $menu.m1
    menu $menu.m1 -tearoff 0
    $menu.m1 add command -label [_ "Pre"] -command \
	[namespace code [list Activate_deactivate_intoolbar pre $names]]
    $menu.m1 add command -label [_ "Post"] -command \
	[namespace code [list Activate_deactivate_intoolbar post $names]]
    $menu.m1 add command -label [_ "PrePost"] -command \
	[namespace code [list Activate_deactivate_intoolbar prepost $names]]

    $menu add separator

    $menu add command -label [_ "Up"] -command \
	    [namespace code [list MoveMacroUpDown $w up]]
    $menu add command -label [_ "Down"] -command \
	    [namespace code [list MoveMacroUpDown $w down]]
    $menu add command -label [_ "Top"] -command \
	    [namespace code [list MoveMacroUpDown $w top]]
    $menu add command -label [_ "Bottom"] -command \
	    [namespace code [list MoveMacroUpDown $w bottom]]

    tk_popup $menu $X $Y
}

proc toolbarmacros::_change_recursive_state { w state } {
    catch { $w configure -state $state }
    foreach i [winfo children $w] {
	_change_recursive_state $i $state
    }
}

proc toolbarmacros::_enabledisable_macrodetails { w } {
    variable macroname

    if { $macroname eq "" } {
	_change_recursive_state $w disabled
    } else {
	_change_recursive_state $w normal
    }
}

proc toolbarmacros::TreeCtrlDouble1 { w x y } {
    set id [$w identify $x $y]
    if {[lindex $id 0] eq "item" && [llength [$w curselection]] == 1 } {
	PlayMacro [winfo toplevel $w]
	return -code break
    }
}

proc toolbarmacros::TreeCtrlDrop { w I l } {

    tk_messageBox -message "%w $I -- $l"
}

# what: undefined, simple or complete
proc toolbarmacros::MacrosWindow { { what undefined } { selnames "" } { w .gid.tmacros } } {
    global GidPriv
    variable images
    variable tablelists
    variable iconbuttons
    variable filterentry
    variable macroname
    
    variable macrointoolbar
    variable macroactive
    variable macroprepost
    variable macrogroup
    variable macrodescription
    variable macrocode

    variable _isinit

    if { !$_isinit} {
	::toolbarmacros::Init
    }

    if { $what eq "undefined" } {
	if { [info exists GidPriv(TMacrosWindowGeom)] } {
	    set what [lindex $GidPriv(TMacrosWindowGeom) 4]
	} else { set what simple }
    }

    destroy $w
    InitWindow $w [_ "Macros window"] TMacrosWindowGeom \
	[list toolbarmacros::MacrosWindow $what]
    if { ![winfo exists $w] } return ;# windows disabled || usemorewindows == 0
    
    labelframe $w.f1 -text [_ "macros list"]

    entry $w.f1.e1 -textvariable toolbarmacros::filterentry

    checkbutton $w.f1.cb1 -image [icon_chooser::giveimage kdmconfig-16] \
	-variable toolbarmacros::filterentry_onlygroup -indicatoron 0 \
	-bd 1
    DynamicHelp::register $w.f1.cb1 balloon [_ "Filter only on group name"]
 
    set filterentry ""
    set cmd "[namespace code [list _filter_MacrosList $w]] ;#"
    trace add variable toolbarmacros::filterentry write $cmd
    bind $w.f1.e1 <Destroy> [list trace remove variable toolbarmacros::filterentry \
	    write $cmd]
    trace add variable toolbarmacros::filterentry_onlygroup write $cmd
    bind $w.f1.cb1 <Destroy> [list trace remove variable \
	    toolbarmacros::filterentry_onlygroup write $cmd]

    ScrolledWindow $w.f1.lf -relief sunken -borderwidth 0

    TreeCtrlList $w.f1.lf.lb -width 450 -height 400 -columns [list \
	    4  "#"        left text \
	    21 [_ "Name"]        left imagetext \
	    6  [_ "Accelerators"] left text \
	    4  [_ "In toolbar"]  center image \
	    4  [_ "Active"]  center image \
	    4  [_ "PrePost"]  center image \
	    10 [_ "Group"] left text \
	    25 [_ "Description"] left text \
	    14 [_ "Creation date"] left text \
	    14 [_ "Modification date"] left text]

    $w.f1.lf.lb notify bind $w.f1.lf.lb <Selection> [namespace code \
	    [list _changeMacrosList_selection $w]]
    bind $w.f1.lf.lb <3> [namespace code [list ContextualMacrosList \
		$w.f1.lf.lb %x %y %X %Y]]
    bind $w.f1.lf.lb <Return> [namespace code [list PlayMacro $w]]
    bind $w.f1.lf.lb <Double-1> [namespace code [list TreeCtrlDouble1 $w.f1.lf.lb %x %y]]
    bind $w.f1.lf.lb <Delete> [namespace code [list DeleteMacro $w]]
    bind $w.f1.lf.lb <Button1-Motion> {
	set toolbarmacros::Dragging 1
    }
    bind $w.f1.lf.lb <ButtonRelease-1> {
	unset -nocomplain toolbarmacros::Dragging
    }

    $w.f1.lf.lb notify bind $w.f1.lf.lb <Drag-receive> {
	set w [winfo toplevel %T]
	set row [lindex [%T item rnc %I] 0]
	toolbarmacros::MoveMacroUpDown $w [list pos $row]
    }

#     tablelist::tablelist $w.f1.lf.lb -width 55 -height 20 \
#         -exportselection 0 \
#         -columns [list \
#             2  [_ #]        left \
#             18 [_ "Name"]        left \
#             5  [_ "Accelerators"] left\
#             3  [_ "In toolbar"]  center \
#             3  [_ "Active"]  center \
#             4  [_ "Icon"]        center \
#             8  [_ "Group"] left \
#             20 [_ "Description"] left \
#             12 [_ "Creation date"] left \
#             12 [_ "Modification date"] left] \
#         -labelcommand tablelist::sortByColumn \
#         -background white \
#         -selectbackground navy -selectforeground white \
#         -stretch end -selectmode extended \
#         -highlightthickness 0
    set tablelists($w) $w.f1.lf.lb
    focus $w.f1.lf.lb
    $w.f1.lf setwidget $w.f1.lf.lb

#     $w.f1.lf.lb columnconfigure 0 -sortmode integer
#     foreach i [list 1 2 6] { 
#         $w.f1.lf.lb columnconfigure $i -sortmode dictionary
#     }

#     #bind $w.f1.lf.lb <<TablelistSelect>> [namespace code [list _changeMacrosList_selection $w]]
#     # can use, for version 3.6 -> bodytag
#     bind [$w.f1.lf.lb bodypath] <1> [list focus %W]
#     bind [$w.f1.lf.lb bodypath] <ButtonRelease-1> [namespace code \
#             [list _changeMacrosList_selection_handler $w]]
#     bind [$w.f1.lf.lb bodypath] <KeyPress> [namespace code \
#                 [list _changeMacrosList_selection_handler $w]]
#     bind [$w.f1.lf.lb bodypath] <3> [namespace code [list ContextualMacrosList \
#                 $w.f1.lf.lb %y %X %Y]]
#     bind [$w.f1.lf.lb bodypath] <Return> [namespace code [list PlayMacro $w]]
#     bind [$w.f1.lf.lb bodypath] <Double-1> [namespace code [list PlayMacro $w]]

    # be careful to update these args, specially the index 7
    set record_macro_args [list $w window $w.f1.bbox1 7]

    set cmds [list forward [_ "Run macro"] [namespace code [list PlayMacro $w]] \
	    remove [_ "Delete macro"] [namespace code [list DeleteMacro $w]] \
	    up [_ "Move macro up"] [namespace code [list MoveMacroUpDown $w up]] \
	    down [_ "Move macro down"] [namespace code [list MoveMacroUpDown $w down]] \
	    top [_ "Move macro top"] [namespace code [list MoveMacroUpDown $w top]] \
	    bottom [_ "Move macro bottom"] [namespace code [list MoveMacroUpDown $w bottom]] \
	    filesave [_ "File operations"] [namespace code \
		[list FileOperations $w $w.f1.bbox1]] \
	    record [_ "Record macro"] [namespace code "StartRecording $record_macro_args"] \
	    ]
    
    set bbox [ButtonBox $w.f1.bbox1 -spacing 0 -padx 1 -pady 1 -homogeneous 1]
    foreach "img help cmd" $cmds {
	$bbox add -image $images($img-16) \
	    -highlightthickness 0 -takefocus 0 -relief link -borderwidth 1 -padx 1 -pady 1 \
	    -helptext $help \
	    -command $cmd
    }
    menubutton $w.f1.b05 -text [_ "View"] -image [GetImage arrow_down_vsmall.gif] \
	-compound right -menu $w.f1.b05.m
    menu $w.f1.b05.m -tearoff 0
    $w.f1.b05.m add command -label [_ "View toolbar"] -command TOBegin

    set cmd [list toolbarmacros::_update_toolbar]
    append cmd ";" [list toolbarmacros::SaveMacros]
    $w.f1.b05.m add checkbutton -label [_ "Toolbar one column"] -variable \
	toolbarmacros::preferences(toolbar_one_col) -command $cmd

    Button $w.f1.b1 -image $images(finish-16) -helptext [_ "View macro details"] \
	-command [list toolbarmacros::ToggleDetailsFrame $w] -relief link -bd 1

    grid $w.f1.e1 - $w.f1.cb1 -sticky ew -padx "3 0" -pady 3
    grid $w.f1.lf  - - -sticky nsew
    grid $bbox $w.f1.b05 $w.f1.b1 -sticky w
    grid configure $w.f1.b1 $w.f1.cb1 -sticky ne -padx 3
    
    grid columnconfigure $w.f1 1 -weight 1
    grid rowconfigure $w.f1 1 -weight 1

    set f2 [labelframe $w.f2 -text [_ "macro details"]]
    label $f2.l1 -text [_ "Name"]:

    frame $f2.fn
    set font [concat [font actual [$f2.l1 cget -font]] -weight bold]
    label $f2.fn.cb1 -textvariable toolbarmacros::macroname -font $font \
	-anchor w

    set cmd [namespace code [list _enabledisable_macrodetails $f2]]
    trace add variable toolbarmacros::macroname write "$cmd;#"
    bind $f2.fn.cb1 <Destroy> [list trace remove variable \
	    toolbarmacros::macroname write "$cmd;#"]

    ButtonBox $f2.fn.b0 -spacing 0 -padx 1 -pady 1 -homogeneous 1
    $f2.fn.b0 add -image [icon_chooser::giveimage edit-16] -helptext \
	[_ "Rename macro"] \
	-relief link -borderwidth 1 -padx 1 -pady 1 -command \
	[namespace code [list _rename_copy_macro $w rename]]
    $f2.fn.b0 add -image [icon_chooser::giveimage editcopy-16] -helptext \
	[_ "Copy macro"] \
	-relief link -borderwidth 1 -padx 1 -pady 1 -command \
	[namespace code [list _rename_copy_macro $w copy]]
    grid $f2.fn.cb1 $f2.fn.b0 -sticky nw
    grid configure $f2.fn.cb1 -sticky ew
    grid columnconfigure $f2.fn 0 -weight 1

    label $f2.lg -text [_ "Group"]:
    ComboBox_and_valuesvar $f2.cbg -textvariable toolbarmacros::macrogroup \
	-valuesvariable ::toolbarmacros::groups -editable 1


    label $f2.l2 -text [_ "Accelerators:"]
    label $f2.l22 -textvariable toolbarmacros::macroaccelerators
    label $f2.l3 -text [_ "New accelerator:"]

    frame $f2.fa
    entry $f2.fa.e1 -textvariable toolbarmacros::macronewaccelerator -width 7
    bind $f2.fa.e1 <KeyPress> "[namespace code [list _enter_accelerators_here_do \
		$f2.fa.e1 %K %A]] ; break"
    bind $f2.fa.e1 <Return> [namespace code [list _manage_accelerators_add $f2.fa.e1]]
    $f2.fa.e1 del 0 end
    ButtonBox $f2.fa.b1 -spacing 0 -padx 1 -pady 1 -homogeneous 1
    $f2.fa.b1 add -image $images(ok-16) -helptext [_ "Define new accelerator"] \
	-relief link -borderwidth 1 -padx 1 -pady 1 -command \
	[namespace code [list _manage_accelerators_add $f2.fa.e1]]
    $f2.fa.b1 add -image $images(remove-16) -helptext [_ "Remove accelerator"] \
	-relief link -borderwidth 1 -padx 1 -pady 1 -command \
	[namespace code [list _manage_accelerators_delete $f2.fa.e1]]
    grid $f2.fa.e1 $f2.fa.b1 -sticky nw
    grid configure $f2.fa.e1 -sticky ew
    grid columnconfigure $f2.fa 0 -weight 1
    
    checkbutton $f2.cb2 -variable toolbarmacros::macrointoolbar -text [_ "In toolbar"]
    frame $f2.fb
    label $f2.fb.l4 -text [_ "Icon"]:
    Button $f2.fb.l5 -image $images(macro1-22) -relief link \
	-helptext [_ "Press to choose icon"] -command \
	[list toolbarmacros::ChangeIcon $f2.fb.l5]
    set iconbuttons($w) $f2.fb.l5
    ComboBox $f2.fb.cbpp -textvariable  toolbarmacros::macroprepost -editable 0 \
	-values [list pre post prepost] -width 6
    
    grid $f2.fb.l4 $f2.fb.l5 $f2.fb.cbpp -padx 3

    checkbutton $f2.cb3 -variable toolbarmacros::macroactive -text [_ "Active"]

    set f21 [labelframe $f2.lf -text [_ "description"]]
    ScrolledWindow $f21.lf -relief sunken -borderwidth 0
    text_and_variable $f21.lf.t -width 20 -height 2 -wrap word \
	-variable ::toolbarmacros::macrodescription -bd 1
    $f21.lf setwidget $f21.lf.t
    grid $f21.lf -sticky nsew -padx 3 -pady 3
    grid columnconfigure $f21 0 -weight 1
    grid rowconfigure $f21 0 -weight 1

    set f22 [labelframe $f2.lf1 -text [_ "code"]]
    ScrolledWindow $f22.lf -relief sunken -borderwidth 0
    text_and_variable $f22.lf.t -width 20 -height 4 -wrap none \
	-variable ::toolbarmacros::macrocode -bd 1
    $f22.lf setwidget $f22.lf.t
    Button $f22.b1 -text [_ "Edit"] -helptext [_ "Edit macro command"] \
	-width 10 -command [namespace code EditMacro] -bd 1
    grid $f22.lf -sticky nsew -padx 3 -pady 3
    grid $f22.b1 -pady 3
    grid columnconfigure $f22 0 -weight 1
    grid rowconfigure $f22 0 -weight 1

    set f23 [frame $f2.f3]
    button $f23.accept -text [_ "Apply"] -width 8 -command \
	[namespace code [list update_macrodetails $w]]
    button $f23.cancel -text [_ "Cancel"] -width 8 -command \
	[namespace code [list _changeMacrosList_selection $w no]]
    grid $f23.accept $f23.cancel -padx 3

    grid $f2.l1 $f2.fn -sticky nw
    grid $f2.lg $f2.cbg -sticky nw -padx "0 2" -pady 2
    grid $f2.l2 $f2.l22 -sticky w
    grid $f2.l3 $f2.fa -sticky w
    grid $f2.cb2  $f2.cb3 -sticky w
    grid $f2.fb    -  -sticky w
    grid $f21   -     -sticky nsew -padx 3 -pady 3
    grid $f22   -     -sticky nsew -padx 3 -pady 3
    grid $f23 -     -pady 3
    grid columnconfigure $f2 1 -weight 1
    grid rowconfigure $f2 "6 7" -weight 1

    set color_fondo [CCGetRGB $w [$w cget -background]]
    frame $w.f_but -bd 0 -background [CCColorActivo $color_fondo]
    button $w.f_but.close -text [_ "Close"] -command [list destroy $w]
    SetWidgetsWidthFromText $w.f_but.close
    
    grid $w.f1 -sticky nsew -padx 3 -pady 3
    grid $w.f_but.close -padx 5 -pady 5
    grid $w.f_but -sticky ews
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    set macroname ""
    set macroaccelerators ""
    set macrointoolbar 0
    set macroactive 0
    set macroprepost prepost
    set macrogroup ""
    set macrodescription ""
    set macrocode ""

    bind $w <Escape> [list destroy $w]
    bind $w <Destroy> "+[namespace code [list update_macrodetails $w]]"

    if { $what eq "complete" } {
	ToggleDetailsFrame $w 0
    }
    _update_MacrosList $w "" $selnames
}

proc toolbarmacros::create_images {} {
    variable images

    if { [info exists images(gear-22)] } { return }

    foreach i [list gear-22 edit-22 forward-16 remove-16 up-16 down-16 \
	    finish-16 ok-16 top-16 bottom-16 filesave-16] {
	set images($i) [icon_chooser::giveimage $i]
    }
    set images(record-16) [image create photo  -data {        
	    R0lGODlhEAAQAJEAANnZ2XoLC5SWlP///yH5BAEAAAAALAAAAAAQABAAAAI5hI+py+1/SByC
	    j0DxCf5RfCP4RvFNCD5QfBMh+BTfRAg+UHwRIfgWESEkieAfySf4CCIiBB9Tl4EKADs=
	}]
    set images(record-22) [image create photo  -data {
	    R0lGODlhFgAWANUAANnZ2XoLC3oMDHoNDXgDA4Q+PZagnngAAIVFRJWdm5SYloQ/PpSXlXkH
	    B4ZNTJGGhJSWlHoJCXkEBIM9PZGIhpajoXsPD30ZGX0YGHsQEIEwL5WamJWdmpSVlJKKiJKJ
	    iJKMipWbmf//////////////////////////////////////////////////////////////
	    /////////////////////////////////////////////////////////yH5BAEAAAAALAAA
	    AAAWABYAAAaxQIBwSCwaj8ikcslsOp/QITAgHAIBwiGxCAwIh0IgQDgkEoEB4VAoAAKEQ6IQ
	    GBAOiQMgQDgkAoAB4XBIKBiAAOGQCAwIh8MDIqEAAoTDITAgHA4PiwQDCBAOh8CAcDg8LBIM
	    IEA4JAIDQqGgcXA8FBAgQDgkBgQRieQwmFAqEAiROIxYLpiMZsPpQCBEIhHg+VBADAYEQiQS
	    hwlOCAKBEIlEIhAiFAIBwiGxaDwik8YgADs=
	}]
    set images(recordend-16) [image create photo  -data {
	    R0lGODlhEAAQAJEAANnZ2XoLC5SWlP///yH5BAEAAAAALAAAAAAQABAAAAJIhI+pqxQfg2BT
	    fAyCTfExRAAIjI8hAkBgfAwRAALjY4gAEBgfQwSAwPgYIgAExscQASAwPoYIAIHxMUQINslH
	    IdgkH4XgY1oBADs=
	}]
    set images(recordend-22) [image create photo  -data {
	    R0lGODlhFgAWAJEAANnZ2XoLC5SWlP///yH5BAEAAAAALAAAAAAWABYAAAJShI+py+0Po5yU
	    xMcg+AgUH4PgI1B8DBGCfxQfQ4TgH8XHECH4R/ExRAj+UXwMEYJ/FB9DhOAfxccQIfhH8TFE
	    CD6SfDSCjyQfjeBj6nL7w2hIAQA7
	}]

    set images(macro1-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXcICCOZAkEJiAIglkOpiAIghmMREGIRDCCggAC4mgMBxEU
	    AXEUxigI4ygWAZIgQaGMACgIICCOorEkSbIwJCAI5Eg0S5Iki0ICgkCKIKIZAidBgkIBAAAQ
	    BBAQRaIZDuIpAuIojFEQRoCAgkhqiIIQQSIAQEEcAQIiIqkhgAEExJEUCYiIpIYQg7IECIiI
	    pIYoS5KQGjCSGgIExJEkiamhqoYoS5KQGqqyiLIkiamhKtAiQEAcSZKQGqqyiLIkiamhKoso
	    y1JqqMoCCRAQR7IEGqqyCNMsqcoiTNMMAQA7
	}]
    set images(macro2-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXfICCOZAkEpiAIglkOpSCIgiCQwUgUhEiAgQAAACAIAAAA
	    xnAQQREQR2GAgCgKwjgWAZIgQaGMoyCMo7EkSbIw4wgKAgiII9EsSZIsyigKwjgSzRA4CRIU
	    iigIICAKIkA0w0E8RUAchQEK4igAAAEFkdQQBQEARECOAAERkdQQADiAgDiSIgERkdQQYlCW
	    AAERkdQQZUkSUgNGUkOAgDiSJDE1VNUQZUkSUkNVFlGWJDE1VAVaBAiII0kSUkNVFlGWJDE1
	    VGURZVlKDVVZIAEC4kiWQENVFmGaJVVZhGmaIQA7
	}]
    set images(macro3-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXlICCOZAkEJSiIowAC4jgOJCAIoyAIYzASBSESgUACgjAa
	    A3gQQREQR2GAgDgKwigWAZIgQaGMoiAI42gsSZIszDgCQIIIAABANEuSJIsCAuIoCsJINEPg
	    JEhQKIBAAoIgEs1wEA9YBMRRGIIAAIAgCAABBZHUEAUBAEQAgIIoAgABEZHUEMAAAuJIigRE
	    RFJDiEFZAgRERFJDlCVJSA0YSQ0BAuJIksTUUFVDlCVJSA1VWURZksTUUBVoESAgjiRJSA1V
	    WURZksTUUJVFlGUpNVRlgQQIiCNZAg1VWYRpllRlEaZphgA7
	}]
    set images(macro4-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXhICCOZAkE5igIgjkOpigIgkAGI1EQIhGAgCgAgiCKxnAQ
	    QREQR2GIgigIoigWAQImSFAoACAAACAIICCOxpIkycKMwigIo0g0S5IkiyKCgjiOAEA0Q+Ak
	    SFAoICCOgjASzXAQTxEQR2GMoiCIBBREUkOABQEARAAAACiII0REUkMAAwiIIykSEBFJDSEG
	    ZQkQEBFJDVGWJCE1YCQ1BAiII0kSU0NVDVGWJCE1VGURZUkSU0NVoEWAgDiSJCE1VGURZUkS
	    U0NVFlGWpdRQlQUSICCOZAk0VGURpllSlUWYphkCADs=
	}]
    set images(macro5-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXjICCOZAkEJiiIIwiI4ziQIiiIIwAAQAAAAFEQAEAEgAAC
	    4jgaw0EERUAchQEIJDkWAZIgQaGIoCCKICCOgLEkSbIw5CgIAkk0S5Iki0KSgjCCRDMEToIE
	    hQIIICCKgiASzXAQTxEQR2EIwigIAkBAQSSBDVEQAEAEACiIIgAQEBFJDQEMICCOpEhARCQ1
	    hBiUJUBARCQ1RFmShNSAkdQQICCOJElMDVU1RFmShNRQlUWUJUlMDVWBFgEC4kiShNRQlUWU
	    JUlMDVVZRFmWUkNVFkiAgDiSJdBQlUWYZklVFmGaZggAOw==
	}]
    set images(macro6-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXzICCOZAkEJgAKoggC4jgOpCgIo0AGI1EQIhEIAkkCxnAQ
	    AVgExFEYggAIggAC4ggUAZIgQaEAgiCIgkCKxpIkycKIAigAAAAIAgAAANEsSZIsCgAIAgAA
	    gCAAIAAQzRA4CRIUCiAIAAAAggAARDMcxFOAAXEUBiAIACAIAEBAQSQ1REEAABEAgAAKggAA
	    BEREUkMAAwiIIykSEBFJDSEGZQkQEBFJDVGWJCE1YCQ1BAiII0kSU0NVDVGWJCE1VGURZUkS
	    U0NVoEWAgDiSJCE1VGURZUkSU0NVFlGWpdRQlQUSICCOZAk0VGURpllSlUWYphkCADs=
	}]
    set images(macro7-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXnICCOZAkEJSiIowAC4jgOJAAK4igAAAAEAAAQBQEARCAA
	    AACIAAAABjgcRFAExFEYICCKwjgCRYAkSFAooygI4wgYS5IkCzOOIAEExBEgmiVJkkUhAUEg
	    RaIZAidBgkIhBRAQR6IZDuIpAuIoDFEQRpGAgkhqiIIQQSIAAAEERJGAiEhqCGAYx3EkICKS
	    GkIMxnEUCRAiIqkhQEAcSZKQmkhqiLIkiamhqoYoS5KQGqoCLQIExJEkiamhKosoS5KQGqqy
	    iLIkiamhKtAiQEAcyVJqqMoiyrIEGqqyiLIsRaqyiLIsR6AQADs=
	}]
    set images(macro8-22) [image create photo  -data {
	    R0lGODlhFgAWAMQAANnZ2dzaTEtFvfz6BAQCBPz+BExKBMzOTPz+rPz+3ISCBPTyhISCLISC
	    hPz+xOTiVPz+/MTCxKSipKyqrExOTDw+PDQyNP//////////////////////////////////
	    /yH5BAEAAAAALAAAAAAWABYAAAXpICCOZAkEJiiIIwiI4ziQgCCQgjAGI1EQIhEIAikIozEc
	    IBEUAXEUhiAAAAAIAgiIYhEgCRIUiggK4ggC4mgsSZIsjCgIIyAII0A0S5IkiyIKAgiIgCCM
	    RDMEToIEhQIIwggIgkg0w0E8YBEQR2EIAgAAgCAABBREUkMUBAAQAQAK4ggQEBFJDQEMICCO
	    pEhARCQ1hBiUJUBARCQ1RFmShNSAkdQQICCOJElMDVU1RFmShNRQlUWUJUlMDVWBFgEC4kiS
	    hNRQlUWUJUlMDVVZRFmWUkNVFkiAgDiSJdBQlUWYZklVFmGaZggAOw==
	}]
    set images(transparent-16) [image create photo  -data {
	    R0lGODlhEAAQAJEAANnZ2f///////////yH5BAEAAAAALAAAAAAQABAAAAIOhI+py+0Po5y0
	    2ouzPqQAOw==
	}]
    set images(pre-24) [image create photo  -data {
	    R0lGODlhGAAYAKIAANnZ2QAAAAAA9//7//cA9/cA//8A9////yH5BAEAAAAALAAAAAAYABgA
	    AAO3CLrc/hCiUARdVhTdjcBA0FUU1RlCIqoJjEBQFRGUQVKtGZwhEkAAEBEcIiIcXS5B0Igg
	    JMPRZSUcAYDACMLRZSYcAYDACBEcXV7CEQRdFUHC0WUSBF0GQcLR3RIEXVZBwtFVEgRdXkHC
	    0R0TBF1mQcLRJUHQ5RYcXRIEXW7B0SVB0OUWJBxVGkHQZRYkXRpB0OUWHN2IQNBlFhwU1YhA
	    0OUNjEBQjQgEXd7ACARdbsdQBF1u/4YSADs=
	}]
    set images(post-24) [image create photo  -data {
	    R0lGODlhGAAYAMQAANnZ2QAAAJwAAOcAAP+OAP+WAP84AP8AAJQAAP/nAP+mAAAA9+//AP/f
	    AAD3/wD/AABh/wD/zgB5/wAAtQAA/wAQ7wAY9wBR/wAA5wAArQAI5wAo/wCO////////////
	    /yH5BAEAAAAALAAAAAAYABgAAAX/ICCOQBAEJEmSJBAEgSCQJEmOQRAMw0CSJNIIRBTGMSAC
	    CIgjWQKJohDHMQyIaYoL0zSNQhzDIJgmaHGeh2EaxTiOQwABcSQXyHGi50mUhiCOQyDJBYIk
	    SHKcB2QYJTmOQQAAAAiCYJooCoIkx3meJiEM8BgEAACCIJimqaKsC3Kc50mUwjgGAQDAQFyW
	    aZomDIIc52EahTgGBATEEViWaZowCHKih2kU4xgEkqSmKZsgEHKih2mIY0BAQBzHZZkmCJKc
	    6EkU4xgEkhSnqaIgyQEjpiGMAwTEkRSnSbMuyXkYhShLcpqwDZKipynLEgCaJmzjnAgExJEs
	    xSnDMEhyGUzTnCaMgkzTBKYJ20zTNAJNICCOZGmeZAgAOw==
	}]
    set images(prepost-24) [image create photo  -data {
	    R0lGODlhGAAYAMQAANnZ2QAAAAAA9/cA9//7//cA/5wAAP8AAOcAAP+OAP84AP/nAP+mAP+W
	    AAD/AO//AP/fAJQAAAD3/wD/zgAA/wBh/wB5/wAAtQAQ7wAI5wAY9wBR/wAA5wAo/wCO/wAA
	    rSH5BAEAAAAALAAAAAAYABgAAAX2ICCOZGmeZBAEKFoGQYCiZRAEgoCiojCABDEMggAC4kiW
	    wgASojgQAgiIIzkKBUiIIzEMggAC4kgKxAAS4jgOhAEC4igKxAAS4kgSB2IAAAAEQUAMZUkk
	    yoEYAAAEARgMA0iII7kwjXIgBgAEQSAIw0COzgMxyYFEICCOgCAMA0EQhDQ5D8QoB2KQJEUM
	    A1GBljQ5D5QcSAQC4jgKAkFRlSVNzsIoB2KQpHhdGFVZEjg9UKIcICCOpHhdmbZZkvMwSVmS
	    18V1lTU5UFmWANB1cZ0nTSAgjmQpXh/HVZZkmuZ1cVRlmiZwXVxnmqYR6AIBcSRL8yRDADs=
	}]
    
}

proc toolbarmacros::is_image_in_use { name } {
    upvar #0 toolbarmacros::macrospace::macrosdata macrosdata
    foreach item [array names macrosdata *,Icon] {
	if { [lindex $macrosdata($item) 0] == $name && \
		[lindex $macrosdata($item) 1] == "imported_images" } {
	    return 1
	}
    }
    return 0
} 
